# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import re
import subprocess
import sys
from argparse import ArgumentParser
from datetime import date
from pathlib import Path
from typing import Dict, Iterable, List, Optional

import tomlkit
from tomlkit.toml_document import TOMLDocument

from pontos.errors import PontosError
from pontos.git import Git, GitError, TagSort
from pontos.release.helper import get_git_repository_name
from pontos.terminal import Terminal
from pontos.terminal.null import NullTerminal
from pontos.terminal.rich import RichTerminal

ADDRESS = "https://github.com/"


class ChangelogBuilderError(PontosError):
    """
    An error while building a changelog
    """


class ChangelogBuilder:
    """
    Creates Changelog files from conventional commits using the git log,
    from the latest tag.
    """

    def __init__(
        self,
        *,
        terminal: Terminal,
        config: Path,
        current_version: str,
        next_version: str,
        space: str,
        project: Optional[str] = None,
    ):
        self._terminal = terminal
        self.config: TOMLDocument = tomlkit.parse(
            config.read_text(encoding="utf-8")
        )
        self.project = (
            project if project is not None else get_git_repository_name()
        )
        self.space = space

        self.current_version = current_version
        self.next_version = next_version

    def create_changelog_file(self, output: str) -> Optional[Path]:
        commit_list = self._get_git_log()
        commit_dict = self._sort_commits(commit_list)
        return self._build_changelog_file(commit_dict, output)

    def _get_git_log(self) -> List[str]:
        """Getting the git log for the current branch.

        If a last tag is found, the `git log` is searched up to the
          last tag, found by `git tag | sort -V | tail -1`.
        Else if no tag is found, the `git log` is searched full, assuming
          this is a initial release.

        Returns:
            A list of `git log` entries
        """
        git = Git()
        latest_tag = None
        try:
            latest_tag = git.list_tags(sort=TagSort.VERSION)[-1]
        except (GitError, IndexError):
            self._terminal.warning("No tag found.")

        if latest_tag:
            return git.log(f"{latest_tag}..HEAD", oneline=True)

        return git.log(oneline=True)

    def _sort_commits(self, commits: List[str]) -> Dict[str, List[str]]:
        """Sort the commits by commit type and group them
        in a dict
        ```
        {
            'Added:': [
                'commit 1 [1234567](..)',
                'commit 2 [1234568](..)',
                '...',
            ],
            'Fixed:': [
                ...
            ],
        }
        ```

        Returns
            The dict containing the commit messages"""
        # get the commit types from the toml
        commit_types = self.config.get("commit_types")

        commit_link = f"{ADDRESS}{self.space}/{self.project}/commit/"

        commit_dict = {}
        if commits and len(commits) > 0:
            for commit in commits:
                commit = commit.split(" ", maxsplit=1)
                for commit_type in commit_types:
                    reg = re.compile(
                        rf'{commit_type["message"]}\s?[:|-]', flags=re.I
                    )
                    match = reg.match(commit[1])
                    if match:
                        if commit_type["group"] not in commit_dict:
                            commit_dict[commit_type["group"]] = []

                        # remove the commit tag from commit message
                        cleaned_msg = (
                            commit[1].replace(match.group(0), "").strip()
                        )
                        commit_dict[commit_type["group"]].append(
                            f"{cleaned_msg} [{commit[0]}]"
                            f"({commit_link}{commit[0]})"
                        )
                        self._terminal.info(f"{commit[0]}: {cleaned_msg}")

        if not commit_dict:
            raise ChangelogBuilderError("No conventional commits found.")

        return commit_dict

    def _build_changelog_file(
        self, commit_dict: Dict[str, List[str]], output: str
    ) -> Optional[Path]:
        """
        Building the changelog file with the passed dict.

        Args:
            commit_dict: dict containing sorted commits
            output: File name to write changelog into

        Returns:
            The path to the changelog file or None
        """

        # changelog header
        changelog = (
            "# Changelog\n\n"
            "All notable changes to this project "
            "will be documented in this file.\n\n"
        )
        if self.next_version:
            changelog += (
                f"## [{self.next_version}] - {date.today().isoformat()}\n"
            )
        else:
            changelog += "## [Unreleased]\n\n"

        # changelog entries
        commit_types = self.config.get("commit_types")
        for commit_type in commit_types:
            if commit_type["group"] in commit_dict.keys():
                changelog += f"\n## {commit_type['group']}\n"
                for msg in commit_dict[commit_type["group"]]:
                    changelog += f"* {msg}\n"

        # comparison line (footer)
        pre = "\n[Unreleased]: "
        compare_link = f"{ADDRESS}{self.space}/{self.project}/compare/"
        if self.next_version and self.current_version:
            pre = f"\n[{self.next_version}]: "
            diff = f"{self.current_version}...{self.next_version}"
        elif self.current_version:
            diff = f"\n{self.current_version}...HEAD"
        else:
            diff = "???...HEAD"

        changelog += f"{pre}{compare_link}{diff}"

        if not changelog:
            return None

        changelog_dir: Path = Path.cwd() / self.config.get("changelog_dir")
        changelog_dir.mkdir(parents=True, exist_ok=True)

        output_path = changelog_dir / output
        output_path.write_text(changelog, encoding="utf-8")
        return output_path


def parse_args(args: Iterable[str] = None) -> ArgumentParser:
    parser = ArgumentParser(
        description="Conventional commits utility.",
        prog="pontos-changelog",
    )

    parser.add_argument(
        "--config",
        "-C",
        default="changelog.toml",
        type=Path,
        help="Conventional commits config file (toml), including conventions.",
    )

    parser.add_argument(
        "--project",
        help="The github project",
    )

    parser.add_argument(
        "--space",
        default="greenbone",
        help="User/Team name in github",
    )

    parser.add_argument(
        "--current-version",
        default="greenbone",
        help="Current version before these changes",
    )

    parser.add_argument(
        "--next-version",
        help="The planned release version",
    )

    parser.add_argument(
        "--output",
        "-o",
        default="unreleased.md",
        help="The path to the output file (.md)",
    )

    parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Don't print messages to the terminal",
    )

    parser.add_argument(
        "--log-file",
        dest="log_file",
        type=str,
        help="Activate logging using the given file path",
    )

    return parser.parse_args(args=args)


def main(
    args=None,
) -> None:
    parsed_args = parse_args(args)

    if parsed_args.quiet:
        term = NullTerminal()
    else:
        term = RichTerminal()

    term.bold_info("pontos-changelog")

    with term.indent():
        try:
            changelog_builder = ChangelogBuilder(
                terminal=term,
                config=parsed_args.config,
                project=args.project,
                space=args.space,
                current_version=args.current_version,
                next_version=args.next_version,
            )
            changelog_builder.create_changelog_file(args.output)
        except ChangelogBuilderError as e:
            term.error(str(e))
            sys.exit(1)
        except subprocess.CalledProcessError as e:
            term.error(f'Could not run command "{e.cmd}".')
            term.out(f"Error was: {e.stderr}")
            sys.exit(1)

    sys.exit(0)



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This script prints out all teams existing in the space of the given
organization
"""

from argparse import ArgumentParser, Namespace

from rich.console import Console
from rich.table import Table

from pontos.github.api import GitHubAsyncRESTApi


def add_script_arguments(parser: ArgumentParser) -> None:
    parser.add_argument("organization")


async def github_script(api: GitHubAsyncRESTApi, args: Namespace) -> int:
    table = Table()
    table.add_column("Name")
    table.add_column("Description")
    table.add_column("URL")
    table.add_column("Privacy")

    count = 0
    async for team in api.teams.get_all(args.organization):
        table.add_row(
            team.name,
            team.description,
            team.html_url,
            team.privacy.value,
        )
        count += 1

    console = Console()
    console.print(table)

    print(f"{count} teams.")
    return 0



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import sys
from argparse import Namespace
from pathlib import Path

import httpx

from pontos.github.api import GitHubRESTApi
from pontos.terminal import Terminal


def tag(terminal: Terminal, args: Namespace) -> None:
    """Github release function for argument class to call"""

    args.tag_func(terminal, args)


def create_tag(terminal: Terminal, args: Namespace) -> None:
    """Github create tag function for
    argument class to call"""

    git = GitHubRESTApi(token=args.token)

    try:
        # Create tag
        sha = git.create_tag(
            repo=args.repo,
            tag=args.tag,
            message=args.message,
            git_object=args.git_object,
            name=args.name,
            email=args.email,
            git_object_type=args.git_object_type,
            date=args.date,
        )

        # Create tag reference
        git.create_tag_reference(repo=args.repo, tag=args.tag, sha=sha)

    except httpx.HTTPError as e:
        terminal.error(str(e))
        sys.exit(1)

    terminal.ok("Tag created.")


def release(terminal: Terminal, args: Namespace) -> None:
    """Github release function for argument class to call"""

    args.re_func(terminal, args)


def create_release(terminal: Terminal, args: Namespace) -> None:
    """Github create release function for
    argument class to call"""

    git = GitHubRESTApi(token=args.token)

    try:
        # Check if release exist
        if git.release_exists(repo=args.repo, tag=args.tag):
            terminal.error(f"Release {args.tag} exist.")
            sys.exit(1)

        # Create release
        git.create_release(
            repo=args.repo,
            tag=args.tag,
            body=args.body,
            name=args.name,
            target_commitish=args.target_commitish,
            draft=args.draft,
            prerelease=args.prerelease,
        )
    except httpx.HTTPError as e:
        terminal.error(str(e))
        sys.exit(1)

    terminal.ok("Release created.")


def pull_request(terminal: Terminal, args: Namespace):
    args.pr_func(terminal, args)


def create_pull_request(terminal: Terminal, args: Namespace):
    git = GitHubRESTApi(token=args.token)

    try:
        # check if branches exist
        if not git.branch_exists(repo=args.repo, branch=args.head):
            terminal.error(
                f"Head branch {args.head} is not existing "
                "or authorisation failed."
            )
            sys.exit(1)

        terminal.ok(f"Head branch {args.head} is existing.")

        if not git.branch_exists(repo=args.repo, branch=args.target):
            terminal.error(
                f"Target branch {args.target} is not existing or "
                "authorisation failed."
            )
            sys.exit(1)

        terminal.ok(f"Target branch {args.target} exists.")

        git.create_pull_request(
            repo=args.repo,
            head_branch=args.head,
            base_branch=args.target,
            title=args.title,
            body=args.body,
        )
    except httpx.HTTPError as e:
        terminal.error(str(e))
        sys.exit(1)

    terminal.ok("Pull Request created.")


def update_pull_request(terminal: Terminal, args: Namespace):
    git = GitHubRESTApi(token=args.token)

    try:
        if args.target:
            # check if branches exist
            if not git.branch_exists(repo=args.repo, branch=args.target):
                terminal.error(
                    f"Target branch {args.target} is not existing or "
                    "authorisation failed."
                )
                sys.exit(1)

            terminal.ok(f"Target branch {args.target} exists.")
        git.update_pull_request(
            repo=args.repo,
            pull_request=args.pull_request,
            base_branch=args.target,
            title=args.title,
            body=args.body,
        )
    except httpx.HTTPError as e:
        terminal.error(str(e))
        sys.exit(1)

    terminal.ok("Pull Request updated.")


def file_status(terminal: Terminal, args: Namespace):
    git = GitHubRESTApi(token=args.token)

    try:
        # check if PR is existing
        if not git.pull_request_exists(
            repo=args.repo, pull_request=args.pull_request
        ):
            terminal.error(
                f"PR {args.pull_request} is not existing "
                "or authorisation failed."
            )
            sys.exit(1)
        terminal.ok(f"PR {args.pull_request} exists.")

        file_dict = git.pull_request_files(
            repo=args.repo,
            pull_request=args.pull_request,
            status_list=args.status,
        )
        for status in args.status:
            terminal.info(f"{status.value}:")
            files = [str(f.resolve()) for f in file_dict[status]]

            for file_string in files:
                terminal.print(file_string)

            if args.output:
                args.output.write("\n".join(files) + "\n")

    except httpx.HTTPError as e:
        terminal.error(str(e))
        sys.exit(1)


def labels(terminal: Terminal, args: Namespace):
    git = GitHubRESTApi(token=args.token)

    try:
        # check if PR is existing
        if not git.pull_request_exists(repo=args.repo, pull_request=args.issue):
            terminal.error(
                f"PR {args.issue} is not existing or authorisation failed."
            )
            sys.exit(1)

        terminal.ok(f"PR {args.issue} exists.")

        issue_labels = git.get_labels(
            repo=args.repo,
            issue=args.issue,
        )

        issue_labels.extend(args.labels)

        git.set_labels(repo=args.repo, issue=args.issue, labels=issue_labels)

    except httpx.HTTPError as e:
        terminal.error(str(e))
        sys.exit(1)


def repos(terminal: Terminal, args: Namespace):
    git = GitHubRESTApi(token=args.token)

    try:
        # check if Orga is existing
        if not git.organisation_exists(orga=args.orga):
            terminal.error(
                f"PR {args.orga} is not existing or authorisation failed."
            )
            sys.exit(1)
        terminal.ok(f"Organization {args.orga} exists.")
        orga_json = git.get_repositories(
            orga=args.orga, repository_type=args.type
        )
        if args.path:
            repo_info = Path(args.path)
            with open(repo_info, encoding="utf-8", mode="w") as fp:
                json.dump(orga_json, fp, indent=2)
        else:
            terminal.print(orga_json)

    except httpx.HTTPError as e:
        terminal.error(str(e))
        sys.exit(1)



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# pylint: disable=C0413,W0108

import os
import unittest
from contextlib import redirect_stdout
from io import StringIO
from pathlib import Path
from unittest.mock import MagicMock, patch

import httpx

from pontos import changelog, release
from pontos.release.helper import version


class SignTestCase(unittest.TestCase):
    def setUp(self) -> None:
        os.environ["GITHUB_TOKEN"] = "foo"
        self.valid_gh_release_response = (
            '{"zipball_url": "zip", "tarball_url":'
            ' "tar", "upload_url":"upload"}'
        )

    @patch("pontos.release.sign.shell_cmd_runner")
    @patch("pontos.release.helper.Path", spec=Path)
    @patch("pontos.github.api.api.httpx")
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.changelog", spec=changelog)
    def test_fail_sign_on_invalid_get_response(
        self,
        changelog_mock,
        version_mock,
        requests_mock,
        _path_mock,
        _shell_mock,
    ):
        version_mock.main.return_value = (True, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")

        fake_get = MagicMock()
        fake_get.status_code = 404
        fake_get.text = self.valid_gh_release_response
        requests_mock.get.return_value = fake_get

        args = [
            "sign",
            "--project",
            "foo",
            "--release-version",
            "0.0.1",
        ]

        with redirect_stdout(StringIO()), self.assertRaises(httpx.HTTPError):
            release.main(
                leave=False,
                args=args,
            )

    @patch("pontos.release.sign.shell_cmd_runner")
    @patch("pontos.release.sign.Path", spec=Path)
    @patch("pontos.helper.Path", spec=Path)
    @patch("pontos.github.api.api.httpx")
    @patch("pontos.github.api.release.httpx")
    @patch("pontos.helper.httpx.stream")
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.changelog", spec=changelog)
    def test_fail_sign_on_upload_fail(
        self,
        changelog_mock,
        version_mock,
        stream_mock,
        request_mock,
        request2_mock,
        _path_mock,
        _path2_mock,
        _shell_mock,
    ):
        version_mock.main.return_value = (True, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")

        fake_get = MagicMock()
        fake_get.status_code = 200
        fake_get.text = self.valid_gh_release_response
        fake_post = MagicMock()
        fake_post.status_code = 500
        fake_post.is_success = False
        fake_post.text = self.valid_gh_release_response
        fake_post.raise_for_status.side_effect = httpx.HTTPStatusError(
            "Server Error", request=MagicMock(), response=fake_post
        )
        request_mock.get.return_value = fake_get
        request_mock.post.return_value = fake_post
        request2_mock.get.return_value = fake_get
        request2_mock.post.return_value = fake_post

        response = MagicMock()
        response.iter_bytes.side_effect = [
            [b"foo", b"bar", b"baz"],
            [b"lorem", b"ipsum"],
        ]
        response_stream = MagicMock()
        response_stream.__enter__.return_value = response
        stream_mock.return_value = response_stream

        args = [
            "sign",
            "--project",
            "foo",
            "--release-version",
            "0.0.1",
        ]
        with redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )

        self.assertFalse(released)

    @patch("pontos.release.sign.shell_cmd_runner")
    @patch("pontos.release.sign.Path", spec=Path)
    @patch("pontos.helper.Path", spec=Path)
    @patch("pontos.github.api.api.httpx")
    @patch("pontos.github.api.release.httpx")
    @patch("pontos.helper.httpx.stream")
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.changelog", spec=changelog)
    def test_successfully_sign(
        self,
        changelog_mock,
        version_mock,
        stream_mock,
        request_mock,
        request_mock2,
        _path_mock,
        _path2_mock,
        _shell_mock,
    ):
        version_mock.main.return_value = (True, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")

        fake_get = MagicMock()
        fake_get.status_code = 200
        fake_get.text = self.valid_gh_release_response
        fake_post = MagicMock()
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        request_mock.get.return_value = fake_get
        request_mock.post.return_value = fake_post
        request_mock2.get.return_value = fake_get
        request_mock2.post.return_value = fake_post

        response = MagicMock()
        response.iter_bytes.side_effect = [
            [b"foo", b"bar", b"baz"],
            [b"lorem", b"ipsum"],
        ]
        response_stream = MagicMock()
        response_stream.__enter__.return_value = response
        stream_mock.return_value = response_stream

        args = [
            "sign",
            "--project",
            "bar",
            "--release-version",
            "0.0.1",
        ]
        with redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )
        self.assertTrue(released)



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict, Iterable, Optional, Tuple

import httpx

from pontos.github.api.client import GitHubAsyncREST
from pontos.github.models.branch import BranchProtection

GITHUB_ACTIONS_APP_ID = 15368


def update_from_applied_settings(
    branch_protection: BranchProtection,
    required_status_checks: Optional[Iterable[Tuple[str, str]]] = None,
    require_branches_to_be_up_to_date: Optional[bool] = None,
    enforce_admins: Optional[bool] = None,
    dismissal_restrictions_users: Optional[Iterable[str]] = None,
    dismissal_restrictions_teams: Optional[Iterable[str]] = None,
    dismissal_restrictions_apps: Optional[Iterable[str]] = None,
    dismiss_stale_reviews: Optional[bool] = None,
    require_code_owner_reviews: Optional[bool] = None,
    required_approving_review_count: Optional[int] = None,
    require_last_push_approval: Optional[bool] = None,
    bypass_pull_request_allowances_users: Optional[Iterable[str]] = None,
    bypass_pull_request_allowances_teams: Optional[Iterable[str]] = None,
    bypass_pull_request_allowances_apps: Optional[Iterable[str]] = None,
    restrictions_users: Optional[Iterable[str]] = None,
    restrictions_teams: Optional[Iterable[str]] = None,
    restrictions_apps: Optional[Iterable[str]] = None,
    required_linear_history: Optional[bool] = None,
    allow_force_pushes: Optional[bool] = None,
    allow_deletions: Optional[bool] = None,
    block_creations: Optional[bool] = None,
    required_conversation_resolution: Optional[bool] = None,
    lock_branch: Optional[bool] = None,
    allow_fork_syncing: Optional[bool] = None,
    required_signatures: Optional[bool] = None,
) -> Dict[str, Any]:
    """
    Return keyword arguments for update_protection_rules by merging existing
    settings with desired updated values.
    """
    kwargs = {}
    if enforce_admins is not None:
        kwargs["enforce_admins"] = enforce_admins
    elif branch_protection.enforce_admins:
        kwargs["enforce_admins"] = branch_protection.enforce_admins.enabled
    else:
        kwargs["enforce_admins"] = None

    if required_linear_history is not None:
        kwargs["required_linear_history"] = required_linear_history
    elif branch_protection.required_linear_history:
        kwargs[
            "required_linear_history"
        ] = branch_protection.required_linear_history.enabled
    else:
        kwargs["required_linear_history"] = None

    if allow_force_pushes is not None:
        kwargs["allow_force_pushes"] = allow_force_pushes
    elif branch_protection.allow_force_pushes:
        kwargs[
            "allow_force_pushes"
        ] = branch_protection.allow_force_pushes.enabled
    else:
        kwargs["allow_force_pushes"] = None

    if allow_deletions is not None:
        kwargs["allow_deletions"] = allow_deletions
    elif branch_protection.allow_deletions:
        kwargs["allow_deletions"] = branch_protection.allow_deletions.enabled
    else:
        kwargs["allow_deletions"] = None

    if required_conversation_resolution is not None:
        kwargs[
            "required_conversation_resolution"
        ] = required_conversation_resolution
    elif branch_protection.required_conversation_resolution:
        kwargs[
            "required_conversation_resolution"
        ] = branch_protection.required_conversation_resolution.enabled
    else:
        kwargs["required_conversation_resolution"] = None

    if block_creations is not None:
        kwargs["block_creations"] = block_creations
    elif branch_protection.block_creations:
        kwargs["block_creations"] = branch_protection.block_creations.enabled
    else:
        kwargs["block_creations"] = None

    if lock_branch is not None:
        kwargs["lock_branch"] = lock_branch
    elif branch_protection.lock_branch:
        kwargs["lock_branch"] = branch_protection.lock_branch.enabled
    else:
        kwargs["lock_branch"] = None

    if allow_fork_syncing is not None:
        kwargs["allow_fork_syncing"] = allow_fork_syncing
    elif branch_protection.allow_fork_syncing:
        kwargs[
            "allow_fork_syncing"
        ] = branch_protection.allow_fork_syncing.enabled
    else:
        kwargs["allow_fork_syncing"] = None
    if required_signatures is not None:
        kwargs["required_signatures"] = required_signatures
    elif branch_protection.required_signatures:
        kwargs[
            "required_signatures"
        ] = branch_protection.required_signatures.enabled
    else:
        kwargs["required_signatures"] = None

    existing_required_status_checks = branch_protection.required_status_checks
    if existing_required_status_checks:
        kwargs[
            "require_branches_to_be_up_to_date"
        ] = existing_required_status_checks.strict
        if existing_required_status_checks.checks is not None:
            kwargs["required_status_checks"] = [
                (c.context, c.app_id)
                for c in existing_required_status_checks.checks
            ]
    if required_status_checks is not None:
        kwargs["required_status_checks"] = required_status_checks
    if require_branches_to_be_up_to_date:
        kwargs[
            "require_branches_to_be_up_to_date"
        ] = require_branches_to_be_up_to_date

    required_pull_request_reviews = (
        branch_protection.required_pull_request_reviews
    )
    if required_pull_request_reviews:
        dismissal_restrictions = (
            required_pull_request_reviews.dismissal_restrictions
        )
        if dismissal_restrictions:
            if dismissal_restrictions.users:
                kwargs["dismissal_restrictions_users"] = [
                    u.login for u in dismissal_restrictions.users
                ]
            else:
                kwargs["dismissal_restrictions_users"] = []
            if dismissal_restrictions.teams:
                kwargs["dismissal_restrictions_teams"] = [
                    t.slug for t in dismissal_restrictions.teams
                ]
            else:
                kwargs["dismissal_restrictions_teams"] = []
            if dismissal_restrictions.apps:
                kwargs["dismissal_restrictions_apps"] = [
                    t.slug for t in dismissal_restrictions.apps
                ]
            else:
                kwargs["dismissal_restrictions_apps"] = []

        kwargs[
            "dismiss_stale_reviews"
        ] = required_pull_request_reviews.dismiss_stale_reviews
        kwargs[
            "require_code_owner_reviews"
        ] = required_pull_request_reviews.require_code_owner_reviews
        kwargs[
            "required_approving_review_count"
        ] = required_pull_request_reviews.required_approving_review_count
        kwargs[
            "require_last_push_approval"
        ] = required_pull_request_reviews.require_last_push_approval

        bypass_pull_request_allowances = (
            required_pull_request_reviews.bypass_pull_request_allowances
        )
        if bypass_pull_request_allowances:
            if bypass_pull_request_allowances.users is not None:
                kwargs["bypass_pull_request_allowances_users"] = [
                    u.login for u in bypass_pull_request_allowances.users
                ]
            else:
                kwargs["bypass_pull_request_allowances_users"] = []
            if bypass_pull_request_allowances.teams is not None:
                kwargs["bypass_pull_request_allowances_teams"] = [
                    t.slug for t in bypass_pull_request_allowances.teams
                ]
            else:
                kwargs["bypass_pull_request_allowances_teams"] = []
            if bypass_pull_request_allowances.apps is not None:
                kwargs["bypass_pull_request_allowances_apps"] = [
                    a.slug for a in bypass_pull_request_allowances.apps
                ]
            else:
                kwargs["bypass_pull_request_allowances_apps"] = []

    existing_restrictions = branch_protection.restrictions
    if existing_restrictions:
        if existing_restrictions.users is not None:
            kwargs["restrictions_users"] = [
                u.login for u in existing_restrictions.users
            ]
        else:
            kwargs["restrictions_users"] = []
        if existing_restrictions.teams is not None:
            kwargs["restrictions_teams"] = [
                t.slug for t in existing_restrictions.teams
            ]
        else:
            kwargs["restrictions_teams"] = []
        if existing_restrictions.apps is not None:
            kwargs["restrictions_apps"] = [
                a.slug for a in existing_restrictions.apps
            ]
        else:
            kwargs["restrictions_apps"] = []

    if dismissal_restrictions_users is not None:
        kwargs["dismissal_restrictions_users"] = list(
            dismissal_restrictions_users
        )
    if dismissal_restrictions_teams is not None:
        kwargs["dismissal_restrictions_teams"] = list(
            dismissal_restrictions_teams
        )
    if dismissal_restrictions_apps is not None:
        kwargs["dismissal_restrictions_apps"] = list(
            dismissal_restrictions_apps
        )
    if bypass_pull_request_allowances_users is not None:
        kwargs["bypass_pull_request_allowances_users"] = list(
            bypass_pull_request_allowances_users
        )
    if bypass_pull_request_allowances_teams is not None:
        kwargs["bypass_pull_request_allowances_teams"] = list(
            bypass_pull_request_allowances_teams
        )
    if bypass_pull_request_allowances_apps is not None:
        kwargs["bypass_pull_request_allowances_apps"] = list(
            bypass_pull_request_allowances_apps
        )
    if dismiss_stale_reviews is not None:
        kwargs["dismiss_stale_reviews"] = dismiss_stale_reviews
    if require_code_owner_reviews is not None:
        kwargs["require_code_owner_reviews"] = require_code_owner_reviews
    if required_approving_review_count is not None:
        kwargs[
            "required_approving_review_count"
        ] = required_approving_review_count
    if require_last_push_approval is not None:
        kwargs["require_last_push_approval"] = require_last_push_approval

    if restrictions_users is not None:
        kwargs["restrictions_users"] = restrictions_users
    if restrictions_teams is not None:
        kwargs["restrictions_teams"] = restrictions_teams
    if restrictions_apps is not None:
        kwargs["restrictions_apps"] = restrictions_apps

    return kwargs


class GitHubAsyncRESTBranches(GitHubAsyncREST):
    async def exists(self, repo: str, branch: str) -> bool:
        """
        Check if a single branch in a repository exists

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Branch name to check
        """
        api = f"/repos/{repo}/branches/{branch}"
        response = await self._client.get(api)
        return response.is_success

    async def delete(self, repo: str, branch: str) -> None:
        """
        Delete a branch on GitHub

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Branch to be deleted

        Raises:
            HTTPStatusError if the request was invalid
        """
        api = f"/repos/{repo}/git/refs/{branch}"
        response = await self._client.delete(api)
        response.raise_for_status()

    async def protection_rules(
        self, repo: str, branch: str
    ) -> BranchProtection:
        """
        Get branch protection rules for a specific repository
        branch

        https://docs.github.com/en/rest/branches/branch-protection#get-branch-protection

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Get protection rules for this branch

        Raises:
            HTTPStatusError if the request was invalid
        """
        api = f"/repos/{repo}/branches/{branch}/protection"
        response = await self._client.get(api)
        response.raise_for_status()
        return BranchProtection.from_dict(response.json())

    async def update_protection_rules(
        self,
        repo: str,
        branch: str,
        *,
        required_status_checks: Optional[Iterable[Tuple[str, str]]] = None,
        require_branches_to_be_up_to_date: Optional[bool] = None,
        enforce_admins: Optional[bool] = None,
        dismissal_restrictions_users: Optional[Iterable[str]] = None,
        dismissal_restrictions_teams: Optional[Iterable[str]] = None,
        dismissal_restrictions_apps: Optional[Iterable[str]] = None,
        dismiss_stale_reviews: Optional[bool] = None,
        require_code_owner_reviews: Optional[bool] = None,
        required_approving_review_count: Optional[int] = None,
        require_last_push_approval: Optional[bool] = None,
        bypass_pull_request_allowances_users: Optional[Iterable[str]] = None,
        bypass_pull_request_allowances_teams: Optional[Iterable[str]] = None,
        bypass_pull_request_allowances_apps: Optional[Iterable[str]] = None,
        restrictions_users: Optional[Iterable[str]] = None,
        restrictions_teams: Optional[Iterable[str]] = None,
        restrictions_apps: Optional[Iterable[str]] = None,
        required_linear_history: Optional[bool] = None,
        allow_force_pushes: Optional[bool] = None,
        allow_deletions: Optional[bool] = None,
        block_creations: Optional[bool] = None,
        required_conversation_resolution: Optional[bool] = None,
        lock_branch: Optional[bool] = None,
        allow_fork_syncing: Optional[bool] = None,
        required_signatures: Optional[bool] = None,
    ) -> BranchProtection:
        """
        Update or create branch protection rules for a specific repository
        branch.

        https://docs.github.com/en/rest/branches/branch-protection#update-branch-protection

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Get protection rules for this branch
            required_status_checks: An iterable of status checks to require in
                order to merge into this branch. Contains tuples of the name of
                the required check and the ID of the GitHub App that must
                provide this check. Set this App ID to None to automatically
                select the GitHub App that has recently provided this check
            require_branches_to_be_up_to_date: Require branches to be up to date
                before merging.
            enforce_admins: Enforce all configured restrictions for
                administrators.
            dismissal_restrictions_users: Specify which users can dismiss pull
                request reviews.
            dismissal_restrictions_teams: Specify which teams can dismiss pull
                request reviews
            dismissal_restrictions_apps: Specify which apps can dismiss pull
                request reviews
            dismiss_stale_reviews: Set to True if you want to automatically
                dismiss approving reviews when someone pushes a new commit.
            require_code_owner_reviews: Blocks merging pull requests until code
                owners review them.
            required_approving_review_count: Specify the number of reviewers
                required to approve pull requests. Use a number between 1 and 6
                or 0 to not require reviewers.
            require_last_push_approval: Whether someone other than the person
                who last pushed to the branch must approve this pull request.
            bypass_pull_request_allowances_users: The list of user logins
                allowed to bypass pull request requirements.
            bypass_pull_request_allowances_teams: The list of team slugs allowed
                to bypass pull request requirements.
            bypass_pull_request_allowances_apps: The list of app slugs allowed
                to bypass pull request requirements.
            restrictions_users: Restrict users who can push to the protected
                branch.
            restrictions_teams: Restrict teams which can push to the protected
                branch.
            restrictions_apps: Restrict apps which can push to the protected
                branch.
            required_linear_history: Enforce a linear commit Git history.
            allow_force_pushes: Permit force pushes to the protected branch by
                anyone with write access to the repository
            allow_deletions: Allow deletion of the protected branch by anyone
                with write access to the repository.
            block_creations: If set to True, the restrictions branch protection
                settings which limits who can push will also block pushes which
                create new branches, unless the push is initiated by a user,
                team, or app which has the ability to push.
            required_conversation_resolution: Require all conversations on code
                to be resolved before a pull request can be merged into a branch
                that matches this rule.
            lock_branch: Whether to set the branch as read-only. If this is
                True, users will not be able to push to the branch.
            allow_fork_syncing: Whether users can pull changes from upstream
                when the branch is locked. Set to True to allow fork syncing.
                Set to False to prevent fork syncing.
            required_signature: True to require signed commits on a branch.

        Raises:
            HTTPStatusError if the request was invalid
        """
        api = f"/repos/{repo}/branches/{branch}/protection"
        data = {
            "enforce_admins": None,
            "required_status_checks": None,
            "required_pull_request_reviews": None,
            "restrictions": None,
        }

        if enforce_admins is not None:
            data["enforce_admins"] = enforce_admins
        if required_linear_history is not None:
            data["required_linear_history"] = required_linear_history
        if allow_force_pushes is not None:
            data["allow_force_pushes"] = allow_force_pushes
        if allow_deletions is not None:
            data["allow_deletions"] = allow_deletions
        if block_creations is not None:
            data["block_creations"] = block_creations
        if required_conversation_resolution is not None:
            data[
                "required_conversation_resolution"
            ] = required_conversation_resolution
        if lock_branch is not None:
            data["lock_branch"] = lock_branch
        if allow_fork_syncing is not None:
            data["allow_fork_syncing"] = allow_fork_syncing

        if require_branches_to_be_up_to_date is not None:
            status_checks = data.get("required_status_checks") or {}
            # checks must be set if strict is set
            status_checks["checks"] = []
            status_checks["strict"] = require_branches_to_be_up_to_date
            data["required_status_checks"] = status_checks

        if required_status_checks is not None:
            status_checks = data.get("required_status_checks") or {}
            checks = []

            for context, app_id in required_status_checks:
                check = {"context": context}
                if app_id:
                    check["app_id"] = app_id

                checks.append(check)

            status_checks["checks"] = checks
            data["required_status_checks"] = status_checks

        if restrictions_users is not None:
            restrictions = data.get("restrictions") or {}
            # teams must be set too if users are set
            r_teams = restrictions.get("teams", [])
            restrictions["teams"] = r_teams
            restrictions["users"] = list(restrictions_users)
            data["restrictions"] = restrictions
        if restrictions_teams is not None:
            restrictions = data.get("restrictions") or {}
            # users must be set too if teams are set
            r_users = restrictions.get("users", [])
            restrictions["users"] = r_users
            restrictions["teams"] = list(restrictions_teams)
            data["restrictions"] = restrictions
        if restrictions_apps is not None:
            restrictions = data.get("restrictions") or {}
            restrictions["apps"] = list(restrictions_apps)
            data["restrictions"] = restrictions

        if dismiss_stale_reviews is not None:
            required_pull_request_reviews = (
                data.get("required_pull_request_reviews") or {}
            )
            required_pull_request_reviews[
                "dismiss_stale_reviews"
            ] = dismiss_stale_reviews
            data[
                "required_pull_request_reviews"
            ] = required_pull_request_reviews
        if require_code_owner_reviews is not None:
            required_pull_request_reviews = (
                data.get("required_pull_request_reviews") or {}
            )
            required_pull_request_reviews[
                "require_code_owner_reviews"
            ] = require_code_owner_reviews
            data[
                "required_pull_request_reviews"
            ] = required_pull_request_reviews
        if required_approving_review_count is not None:
            required_pull_request_reviews = (
                data.get("required_pull_request_reviews") or {}
            )
            required_pull_request_reviews[
                "required_approving_review_count"
            ] = required_approving_review_count
            data[
                "required_pull_request_reviews"
            ] = required_pull_request_reviews
        if require_last_push_approval is not None:
            required_pull_request_reviews = (
                data.get("required_pull_request_reviews") or {}
            )
            required_pull_request_reviews[
                "require_last_push_approval"
            ] = require_last_push_approval
            data[
                "required_pull_request_reviews"
            ] = required_pull_request_reviews
        if dismissal_restrictions_users is not None:
            required_pull_request_reviews = (
                data.get("required_pull_request_reviews") or {}
            )
            dismissal_restrictions = required_pull_request_reviews.get(
                "dismissal_restrictions", {}
            )
            dismissal_restrictions["users"] = list(dismissal_restrictions_users)
            required_pull_request_reviews[
                "dismissal_restrictions"
            ] = dismissal_restrictions
            data[
                "required_pull_request_reviews"
            ] = required_pull_request_reviews
        if dismissal_restrictions_teams is not None:
            required_pull_request_reviews = (
                data.get("required_pull_request_reviews") or {}
            )
            dismissal_restrictions = required_pull_request_reviews.get(
                "dismissal_restrictions", {}
            )
            dismissal_restrictions["teams"] = list(dismissal_restrictions_teams)
            required_pull_request_reviews[
                "dismissal_restrictions"
            ] = dismissal_restrictions
            data[
                "required_pull_request_reviews"
            ] = required_pull_request_reviews
        if dismissal_restrictions_apps is not None:
            required_pull_request_reviews = (
                data.get("required_pull_request_reviews") or {}
            )
            dismissal_restrictions = required_pull_request_reviews.get(
                "dismissal_restrictions", {}
            )
            dismissal_restrictions["apps"] = list(dismissal_restrictions_apps)
            required_pull_request_reviews[
                "dismissal_restrictions"
            ] = dismissal_restrictions
            data[
                "required_pull_request_reviews"
            ] = required_pull_request_reviews
        if bypass_pull_request_allowances_users is not None:
            required_pull_request_reviews = (
                data.get("required_pull_request_reviews") or {}
            )
            bypass_pull_request_allowances = required_pull_request_reviews.get(
                "bypass_pull_request_allowances", {}
            )
            bypass_pull_request_allowances["users"] = list(
                bypass_pull_request_allowances_users
            )
            required_pull_request_reviews[
                "bypass_pull_request_allowances"
            ] = bypass_pull_request_allowances
            data[
                "required_pull_request_reviews"
            ] = required_pull_request_reviews
        if bypass_pull_request_allowances_teams is not None:
            required_pull_request_reviews = (
                data.get("required_pull_request_reviews") or {}
            )
            bypass_pull_request_allowances = required_pull_request_reviews.get(
                "bypass_pull_request_allowances", {}
            )
            bypass_pull_request_allowances["teams"] = list(
                bypass_pull_request_allowances_teams
            )
            required_pull_request_reviews[
                "bypass_pull_request_allowances"
            ] = bypass_pull_request_allowances
            data[
                "required_pull_request_reviews"
            ] = required_pull_request_reviews
        if bypass_pull_request_allowances_apps is not None:
            required_pull_request_reviews = (
                data.get("required_pull_request_reviews") or {}
            )
            bypass_pull_request_allowances = required_pull_request_reviews.get(
                "bypass_pull_request_allowances", {}
            )
            bypass_pull_request_allowances["apps"] = list(
                bypass_pull_request_allowances_apps
            )
            required_pull_request_reviews[
                "bypass_pull_request_allowances"
            ] = bypass_pull_request_allowances
            data[
                "required_pull_request_reviews"
            ] = required_pull_request_reviews
        if required_signatures is not None:
            data["required_signatures"] = required_signatures

        response = await self._client.put(api, data=data)
        response.raise_for_status()
        return BranchProtection.from_dict(response.json())

    async def delete_protection_rules(self, repo: str, branch: str) -> None:
        """
        Delete branch protection rules for a specific repository branch

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Delete protection rules for this branch

        Raises:
            HTTPStatusError if the request was invalid
        """
        api = f"/repos/{repo}/branches/{branch}/protection"
        response = await self._client.delete(api)
        response.raise_for_status()

    async def set_enforce_admins(
        self, repo: str, branch: str, *, enforce_admins: bool
    ) -> None:
        """
        Enable/disable enforce admin branch protection rule for a specific
        repository branch.

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Delete protection rules for this branch
            enforce_admins: True to enable. False do disable.

        Raises:
            HTTPStatusError if the request was invalid
        """
        api = f"/repos/{repo}/branches/{branch}/protection/enforce_admins"
        if enforce_admins:
            response = await self._client.post(api)
        else:
            response = await self._client.delete(api)

        response.raise_for_status()

    async def set_required_signatures(
        self, repo: str, branch: str, *, required_signatures: bool
    ) -> None:
        """
        Enable/disable required signed commits for a repository branch.

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Delete protection rules for this branch
            required_signature: True to enable. False do disable.

        Raises:
            HTTPStatusError if the request was invalid
        """
        api = f"/repos/{repo}/branches/{branch}/protection/required_signatures"
        if required_signatures:
            response = await self._client.post(api)
        else:
            response = await self._client.delete(api)

        response.raise_for_status()

    async def update_required_status_checks(
        self,
        repo: str,
        branch: str,
        *,
        required_status_checks: Optional[Iterable[Tuple[str, str]]] = None,
        require_branches_to_be_up_to_date: Optional[bool] = None,
    ) -> None:
        """
        Update required status check branch protection rules of a repository
        branch.

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Delete protection rules for this branch
            required_status_checks: An iterable of status checks to require in
                order to merge into this branch. Contains tuples of the name of
                the required check and the ID of the GitHub App that must
                provide this check. Set this App ID to None to automatically
                select the GitHub App that has recently provided this check
            require_branches_to_be_up_to_date: Require branches to be up to date
                before merging.

        Raises:
            HTTPStatusError if the request was invalid
        """
        api = (
            f"/repos/{repo}/branches/{branch}/protection/required_status_checks"
        )
        data = {}
        if require_branches_to_be_up_to_date is not None:
            data["strict"] = require_branches_to_be_up_to_date
        if required_status_checks is not None:
            checks = []
            for context, app_id in required_status_checks:
                check = {"context": context}
                if app_id:
                    check["app_id"] = app_id
                checks.append(check)
            data["checks"] = checks

        response = await self._client.patch(api, data=data)
        response.raise_for_status()

    async def remove_required_status_checks(
        self,
        repo: str,
        branch: str,
    ) -> None:
        """
        Remove required status check branch protection rules of a repository
        branch.

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Delete protection rules for this branch

        Raises:
            HTTPStatusError if the request was invalid
        """
        api = (
            f"/repos/{repo}/branches/{branch}/protection/required_status_checks"
        )
        response = await self._client.delete(api)
        response.raise_for_status()


class GitHubRESTBranchMixin:
    def branch_exists(self, repo: str, branch: str) -> bool:
        """
        Check if a single branch in a repository exists

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Branch name to check
        """
        api = f"/repos/{repo}/branches/{branch}"
        response: httpx.Response = self._request(api)
        return response.is_success

    def delete_branch(self, repo: str, branch: str):
        """
        Delete a branch on GitHub

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Branch to be deleted

        Raises:
            HTTPError if the request was invalid
        """
        api = f"/repos/{repo}/git/refs/{branch}"
        response: httpx.Response = self._request(api, request=httpx.delete)
        response.raise_for_status()

    def branch_protection_rules(self, repo: str, branch: str):
        """
        Get branch protection rules for a specific repository
        branch

        Args:
            repo: GitHub repository (owner/name) to use
            branch: Get protection rules for this branch

        Raises:
            HTTPError if the request was invalid
        """
        api = f"/repos/{repo}/branches/{branch}/protection"
        response: httpx.Response = self._request(api)
        return response.json()



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import importlib
from pathlib import Path

import tomlkit

from .helper import (
    VersionError,
    check_develop,
    is_version_pep440_compliant,
    safe_version,
    strip_version,
    versions_equal,
)
from .version import VersionCommand

TEMPLATE = """# pylint: disable=invalid-name

# THIS IS AN AUTOGENERATED FILE. DO NOT TOUCH!

__version__ = "{}"\n"""


# This class is used for Python Version command(s)
class PythonVersionCommand(VersionCommand):
    def __init__(self, *, project_file_path: Path = None) -> None:
        if not project_file_path:
            project_file_path = Path.cwd() / "pyproject.toml"

        if not project_file_path.exists():
            raise VersionError(f"{str(project_file_path)} file not found.")

        self.pyproject_toml = tomlkit.parse(
            project_file_path.read_text(encoding="utf-8")
        )

        if (
            "tool" not in self.pyproject_toml
            or "pontos" not in self.pyproject_toml["tool"]
            or "version" not in self.pyproject_toml["tool"]["pontos"]
        ):
            raise VersionError(
                "[tool.pontos.version] section missing "
                f"in {str(project_file_path)}."
            )

        pontos_version_settings = self.pyproject_toml["tool"]["pontos"][
            "version"
        ]

        try:
            version_file_path = Path(
                pontos_version_settings["version-module-file"]
            )
        except tomlkit.exceptions.NonExistentKey:
            raise VersionError(
                "version-module-file key not set in [tool.pontos.version] "
                f"section of {str(project_file_path)}."
            ) from None

        super().__init__(
            version_file_path=version_file_path,
            project_file_path=project_file_path,
        )

    def _get_version_from_pyproject_toml(self) -> str:
        """
        Return the version information from the [tool.poetry] section of the
        pyproject.toml file. The version may be in non standardized form.
        """

        if (
            "tool" in self.pyproject_toml
            and "poetry" in self.pyproject_toml["tool"]
            and "version" in self.pyproject_toml["tool"]["poetry"]
        ):
            return self.pyproject_toml["tool"]["poetry"]["version"]

        raise VersionError(
            "Version information not found in "
            f"{str(self.project_file_path)} file."
        )

    def _update_version_file(self, new_version: str) -> None:
        """
        Update the version file with the new version
        """
        self.version_file_path.write_text(
            TEMPLATE.format(new_version), encoding="utf-8"
        )

    def _update_pyproject_version(
        self,
        new_version: str,
    ) -> None:
        """
        Update the version in the pyproject.toml file
        """
        pyproject_toml = tomlkit.parse(
            self.project_file_path.read_text(encoding="utf-8")
        )

        if "tool" not in pyproject_toml:
            tool_table = tomlkit.table()
            pyproject_toml["tool"] = tool_table

        if "poetry" not in pyproject_toml["tool"]:
            poetry_table = tomlkit.table()
            # pylint: disable=line-too-long
            # ignore pylint (2.13.9) error: pontos/version/python.py:128:12: E1101: Instance of 'OutOfOrderTableProxy' has no 'add' member (no-member)
            # pylint: disable=no-member
            pyproject_toml["tool"].add("poetry", poetry_table)

        pyproject_toml["tool"]["poetry"]["version"] = new_version

        self.project_file_path.write_text(
            tomlkit.dumps(pyproject_toml), encoding="utf-8"
        )

    def get_current_version(self) -> str:
        version_module_name = self.version_file_path.stem
        module_parts = list(self.version_file_path.parts[:-1]) + [
            version_module_name
        ]
        module_name = ".".join(module_parts)
        try:
            version_module = importlib.import_module(module_name)
        except ModuleNotFoundError:
            raise VersionError(
                f"Could not load version from {module_name}. Import failed."
            ) from None

        return version_module.__version__

    def verify_version(self, version: str) -> None:
        current_version = self.get_current_version()
        if not is_version_pep440_compliant(current_version):
            raise VersionError(
                f"The version {current_version} in "
                f"{str(self.version_file_path)} is not PEP 440 compliant."
            )

        pyproject_version = self._get_version_from_pyproject_toml()

        if pyproject_version != current_version:
            raise VersionError(
                f"The version {pyproject_version} in "
                f"{str(self.project_file_path)} doesn't match the current "
                f"version {current_version}."
            )

        if version != "current":
            provided_version = strip_version(version)
            if provided_version != current_version:
                raise VersionError(
                    f"Provided version {provided_version} does not match the "
                    f"current version {current_version}."
                )

        self._print("OK")

    def update_version(
        self, new_version: str, *, develop: bool = False, force: bool = False
    ) -> None:

        new_version = safe_version(new_version)
        if check_develop(new_version) and develop:
            develop = False
        if develop:
            new_version = f"{new_version}.dev1"

        pyproject_version = self._get_version_from_pyproject_toml()

        if not self.version_file_path.exists():
            self.version_file_path.touch()
        elif not force and versions_equal(
            new_version, self.get_current_version()
        ):
            self._print("Version is already up-to-date.")
            return

        self._update_pyproject_version(new_version=new_version)

        self._update_version_file(new_version=new_version)

        self._print(
            f"Updated version from {pyproject_version} to {new_version}"
        )



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from argparse import Namespace
from pathlib import Path

from pontos import changelog
from pontos.git import Git
from pontos.terminal import Terminal

from .helper import (
    calculate_calendar_version,
    commit_files,
    find_signing_key,
    get_git_repository_name,
    get_last_release_version,
    get_next_patch_version,
    update_version,
)

RELEASE_TEXT_FILE = ".release.md"


def prepare(
    terminal: Terminal,
    args: Namespace,
    **_kwargs,
) -> bool:
    git_tag_prefix: str = args.git_tag_prefix
    git_signing_key: str = (
        args.git_signing_key
        if args.git_signing_key is not None
        else find_signing_key(terminal)
    )
    project: str = (
        args.project if args.project is not None else get_git_repository_name()
    )
    space: str = args.space
    calendar: bool = args.calendar
    patch: bool = args.patch

    if calendar:
        release_version: str = calculate_calendar_version(terminal)
    elif patch:
        release_version: str = get_next_patch_version(terminal)
    else:
        release_version: str = args.release_version

    terminal.info(f"Preparing the release {release_version}")

    # guardian
    git = Git()
    git_tags = git.list_tags()
    git_version = f"{git_tag_prefix}{release_version}"
    if git_version in git_tags:
        terminal.error(f"git tag {git_version} is already taken.")
        sys.exit(1)

    executed, filename = update_version(terminal, release_version)
    if not executed:
        return False

    terminal.ok(f"updated version  in {filename} to {release_version}")

    changelog_bool = True
    if args.conventional_commits:
        last_release_version = get_last_release_version()
        changelog_builder = changelog.ChangelogBuilder(
            terminal=terminal,
            current_version=last_release_version,
            next_version=release_version,
            space=space,
            project=project,
            config=args.cc_config,
        )

        output_file = changelog_builder.create_changelog_file(
            f"v{release_version}.md"
        )
        terminal.ok(f"Created changelog {output_file}")
        commit_msg = f"Changelog created for release to {release_version}"
        commit_files(
            git,
            output_file,
            commit_msg,
            git_signing_key=git_signing_key,
        )
        changelog_bool = False
        # Remove the header for the release text
        changelog_text = output_file.read_text(encoding="utf-8").replace(
            "# Changelog\n\n"
            "All notable changes to this project "
            "will be documented in this file.\n\n",
            "",
        )
    else:
        change_log_path = Path.cwd() / "CHANGELOG.md"
        if args.changelog:
            tmp_path = Path.cwd() / Path(args.changelog)
            if tmp_path.is_file():
                change_log_path = tmp_path
            else:
                terminal.warning(f"{tmp_path} is not a file.")

        # Try to get the unreleased section of the specific version
        updated, changelog_text = changelog.update(
            change_log_path.read_text(encoding="utf-8"),
            release_version,
            git_tag_prefix=git_tag_prefix,
            containing_version=release_version,
        )

        if not updated:
            # Try to get unversioned unrelease section
            updated, changelog_text = changelog.update(
                change_log_path.read_text(encoding="utf-8"),
                release_version,
                git_tag_prefix=git_tag_prefix,
            )

        if not updated:
            terminal.error("No unreleased text found in CHANGELOG.md")
            sys.exit(1)

        change_log_path.write_text(updated, encoding="utf-8")

        terminal.ok("Updated CHANGELOG.md")

    terminal.info("Committing changes")

    commit_msg = f"Automatic release to {release_version}"
    commit_files(
        git,
        filename,
        commit_msg,
        git_signing_key=git_signing_key,
        changelog=changelog_bool,
    )

    git.tag(git_version, gpg_key_id=git_signing_key, message=commit_msg)

    release_text = Path(RELEASE_TEXT_FILE)
    release_text.write_text(changelog_text, encoding="utf-8")

    terminal.warning(
        f"Please verify git tag {git_version}, "
        f"commit and release text in {str(release_text)}"
    )
    terminal.print("Afterwards please execute release")

    return True



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import List, Optional

from pontos.github.models.base import GitHubModel, Team, User
from pontos.github.models.organization import Repository

__all__ = ("PullRequest",)


@dataclass
class CommitUser(GitHubModel):
    name: str
    email: str
    date: str


@dataclass
class Tree(GitHubModel):
    url: str
    sha: str


@dataclass
class CommitVerification(GitHubModel):
    verified: bool
    reason: str
    signature: Optional[str] = None
    payload: Optional[str] = None


@dataclass
class Commit(GitHubModel):
    url: str
    author: CommitUser
    committer: CommitUser
    message: str
    tree: Tree
    comment_count: str
    verification: CommitVerification


@dataclass
class CommitParent(GitHubModel):
    url: str
    sha: str


@dataclass
class PullRequestCommit(GitHubModel):
    url: str
    sha: str
    node_id: str
    html_url: str
    comments_url: str
    commit: Commit
    author: User
    committer: User
    parents: List[CommitParent]


@dataclass
class Label(GitHubModel):
    id: int
    node_id: str
    url: str
    name: str
    description: str
    color: str
    default: bool


@dataclass
class Milestone(GitHubModel):
    url: str
    html_url: str
    labels_url: str
    id: int
    node_id: str
    number: int
    state: str
    title: str
    description: str
    creator: User
    open_issues: int
    closed_issues: int
    created_at: str
    updated_at: str
    closed_at: str
    due_on: str


@dataclass
class PullRequestRef(GitHubModel):
    label: str
    ref: str
    sha: str
    user: User
    repo: Repository


@dataclass
class PullRequest(GitHubModel):
    url: str
    id: int
    node_id: str
    html_url: str
    diff_url: str
    patch_url: str
    issue_url: str
    commits_url: str
    review_comments_url: str
    review_comment_url: str
    comments_url: str
    statuses_url: str
    number: int
    state: str
    locked: bool
    title: str
    user: User
    body: str
    labels: List[Label]
    milestone: Milestone
    active_lock_reason: str
    created_at: str
    updated_at: str
    closed_at: str
    merged_at: str
    merge_commit_sha: str
    assignee: User
    assignees: List[User]
    requested_reviewers: List[User]
    requested_teams: List[Team]
    head: PullRequestRef
    base: PullRequestRef
    author_association: str
    draft: bool
    merged: bool
    mergeable: bool
    rebaseable: bool
    mergeable_state: str
    merged_by: User
    comments: int
    review_comments: int
    maintainer_can_modify: bool
    commits: int
    additions: int
    deletions: int
    changed_files: int
    auto_merge: Optional[bool] = None



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from argparse import Namespace
from pathlib import Path

import httpx

from pontos import changelog
from pontos.git import Git
from pontos.github.api import GitHubRESTApi
from pontos.terminal import Terminal

from .helper import (
    commit_files,
    find_signing_key,
    get_current_version,
    get_git_repository_name,
    get_next_dev_version,
    update_version,
)

RELEASE_TEXT_FILE = ".release.md"


def release(
    terminal: Terminal,
    args: Namespace,
    *,
    token: str,
    **_kwargs,
) -> bool:
    if not token:
        terminal.error(
            "Token is missing. The GitHub token is required to create a "
            "release."
        )
        return False

    project: str = (
        args.project if args.project is not None else get_git_repository_name()
    )
    space: str = args.space
    git_signing_key: str = (
        args.git_signing_key
        if args.git_signing_key is not None
        else find_signing_key(terminal)
    )
    git_remote_name: str = (
        args.git_remote_name if args.git_remote_name is not None else ""
    )
    git_tag_prefix: str = args.git_tag_prefix
    release_version: str = (
        args.release_version
        if args.release_version is not None
        else get_current_version(terminal)
    )
    if not release_version:
        terminal.error("No release version available.")
        return False

    next_version: str = (
        args.next_version
        if args.next_version is not None
        else get_next_dev_version(release_version)
    )

    terminal.info("Pushing changes")

    git = Git()
    git.push(follow_tags=True, remote=git_remote_name)

    terminal.info(f"Creating release for v{release_version}")
    changelog_text: str = Path(RELEASE_TEXT_FILE).read_text(encoding="utf-8")

    github = GitHubRESTApi(token=token)

    git_version = f"{git_tag_prefix}{release_version}"
    repo = f"{space}/{project}"

    try:
        github.create_release(
            repo,
            git_version,
            name=f"{project} {release_version}",
            body=changelog_text,
        )
    except httpx.HTTPError as e:
        terminal.error(str(e))
        return False

    Path(RELEASE_TEXT_FILE).unlink()

    commit_msg = (
        f"Automatic adjustments after release\n\n"
        f"* Update to version {next_version}\n"
    )

    # set to new version add skeleton
    changelog_bool = False
    if not args.conventional_commits:
        change_log_path = Path.cwd() / "CHANGELOG.md"
        if args.changelog:
            tmp_path = Path.cwd() / Path(args.changelog)
            if tmp_path.is_file():
                change_log_path = tmp_path
            else:
                terminal.warning(f"{tmp_path} is not a file.")

        updated = changelog.add_skeleton(
            markdown=change_log_path.read_text(encoding="utf-8"),
            new_version=release_version,
            project_name=project,
            git_tag_prefix=git_tag_prefix,
            git_space=space,
        )
        change_log_path.write_text(updated, encoding="utf-8")
        changelog_bool = True

        commit_msg += f"* Add empty changelog after {release_version}"

    executed, filename = update_version(terminal, next_version, develop=True)
    if not executed:
        return False

    commit_files(
        git,
        filename=filename,
        commit_msg=commit_msg,
        git_signing_key=git_signing_key,
        changelog=changelog_bool,
    )

    # pushing the new tag
    git.push(follow_tags=True, remote=git_remote_name)

    return True



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import datetime
import subprocess
import unittest
from pathlib import Path
from typing import Generator
from unittest.mock import MagicMock, patch

from pontos.git.git import ConfigScope, Git, GitError, exec_git
from pontos.release.helper import (
    calculate_calendar_version,
    find_signing_key,
    get_git_repository_name,
    get_last_release_version,
    get_next_dev_version,
    get_next_patch_version,
    update_version,
)
from pontos.testing import temp_directory, temp_git_repository


@contextlib.contextmanager
def init_test_git_repo() -> Generator[Path, None, None]:
    with temp_git_repository() as repo_path:
        exec_git("remote", "add", "foo", "https://foo.bar/bla.git")
        exec_git("remote", "add", "origin", "https://foo.bar/testrepo.git")
        yield repo_path


class TestHelperFunctions(unittest.TestCase):
    def setUp(self):
        self.shell_cmd_runner = lambda x, cwd=None: subprocess.run(
            x,
            shell=True,
            check=True,
            errors="utf-8",  # use utf-8 encoding for error output
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=cwd,
        )

    def test_get_project_name(self):
        with init_test_git_repo():
            project = get_git_repository_name(remote="foo")
            self.assertEqual(project, "bla")

            project = get_git_repository_name()
            self.assertEqual(project, "testrepo")

    def test_find_signing_key(self):
        terminal = MagicMock()

        with temp_git_repository():
            git = Git()
            git.config(
                "user.signingkey",
                "1234567890ABCEDEF1234567890ABCEDEF123456",
                scope=ConfigScope.LOCAL,
            )
            signing_key = find_signing_key(terminal)
            self.assertEqual(
                signing_key, "1234567890ABCEDEF1234567890ABCEDEF123456"
            )

    def test_find_no_signing_key(self):
        terminal = MagicMock()
        saved_key = None

        git = Git()
        try:
            # save possibly set git signing key from user temporarily
            try:
                saved_key = git.config(
                    "user.signingkey", scope=ConfigScope.GLOBAL
                )
            except GitError:
                saved_key = None

            try:
                git.config("user.signingkey", "", scope=ConfigScope.GLOBAL)
            except subprocess.CalledProcessError as e:
                self.assertEqual(e.returncode, 5)

            with temp_git_repository():
                signing_key = find_signing_key(terminal)
                self.assertEqual(signing_key, "")

        finally:
            # reset the previously saved signing key ...
            if saved_key is not None:
                git.config(
                    "user.signingkey", saved_key, scope=ConfigScope.GLOBAL
                )

    def test_update_version_not_found(self):
        terminal = MagicMock()

        with temp_directory(change_into=True):
            executed, filename = update_version(
                terminal, to="21.4.4", develop=True
            )

        self.assertFalse(executed)
        self.assertEqual(filename, "")

    def test_update_version_no_version_file(self):
        terminal = MagicMock()

        with temp_directory(change_into=True, add_to_sys_path=True) as tmp_dir:
            module_path = tmp_dir / "testrepo"
            module_path.mkdir(parents=True, exist_ok=True)

            init_file = module_path / "__init__.py"
            init_file.touch()

            version_file = module_path / "__version__.py"

            toml = tmp_dir / "pyproject.toml"
            toml.write_text(
                "[tool.poetry]\n"
                'name = "testrepo"\n'
                'version = "21.6.2.dev1"\n\n'
                "[tool.pontos.version]\n"
                'version-module-file = "testrepo/__version__.py"\n',
                encoding="utf-8",
            )

            executed, filename = update_version(
                terminal, to="21.4.4", develop=False
            )

            toml_text = toml.read_text(encoding="utf-8")

            self.assertEqual(filename, "pyproject.toml")
            self.assertTrue(executed)
            self.assertEqual(
                toml_text,
                "[tool.poetry]\n"
                'name = "testrepo"\n'
                'version = "21.4.4"\n\n'
                "[tool.pontos.version]\n"
                'version-module-file = "testrepo/__version__.py"\n',
            )
            self.assertTrue(version_file.exists())
            version_text = version_file.read_text(encoding="utf-8")
            self.assertEqual(
                version_text,
                "# pylint: disable=invalid-name\n\n"
                "# THIS IS AN AUTOGENERATED FILE. DO NOT TOUCH!\n\n"
                '__version__ = "21.4.4"\n',
            )

    @patch("pontos.release.helper.Git", spec=Git)
    def test_get_last_release_version_git(self, _git_interface_mock):
        git_interface = _git_interface_mock.return_value
        git_interface.list_tags.return_value = ["1", "2", "3.55"]
        self.assertEqual(get_last_release_version(), "3.55")

    @patch("pontos.release.helper.Git", spec=Git)
    def test_get_no_release_version_git(self, _git_interface_mock):
        git_interface = _git_interface_mock.return_value
        git_interface.list_tags.return_value = []
        self.assertIsNone(get_last_release_version())


class CalculateHelperVersionTestCase(unittest.TestCase):
    def test_calculate_calendar_versions(self):
        terminal = MagicMock()
        today = datetime.datetime.today()

        filenames = ["pyproject.toml", "CMakeLists.txt"]
        mocks = [
            "pontos.release.helper.PythonVersionCommand",
            "pontos.release.helper.CMakeVersionCommand",
        ]
        current_versions = [
            "21.4.1.dev3",
            f"19.{str(today.month)}.1.dev3",
            f"{str(today.year % 100)}.{str(today.month)}.1.dev3",
            f"{str(today.year % 100)}.{str(today.month)}.1",
        ]
        assert_versions = [
            f"{str(today.year % 100)}.{str(today.month)}.0",
            f"{str(today.year % 100)}.{str(today.month)}.0",
            f"{str(today.year % 100)}.{str(today.month)}.1",
            f"{str(today.year % 100)}.{str(today.month)}.2",
        ]

        with temp_directory(change_into=True) as tmp_path:
            for filename, mock in zip(filenames, mocks):
                for current_version, assert_version in zip(
                    current_versions, assert_versions
                ):
                    proj_file = tmp_path / filename
                    proj_file.touch()
                    with patch(mock) as cmd_mock:
                        # pylint: disable=line-too-long
                        cmd_mock.return_value.get_current_version.return_value = (
                            current_version
                        )

                        release_version = calculate_calendar_version(terminal)
                        self.assertEqual(release_version, assert_version)

                    proj_file.unlink()

    def test_get_next_dev_version(self):
        current_versions = [
            "20.4.1",
            "20.4.1",
            "19.1.2",
            "1.1.1",
            "20.6.1",
        ]
        assert_versions = [
            "20.4.2",
            "20.4.2",
            "19.1.3",
            "1.1.2",
            "20.6.2",
        ]

        for current_version, assert_version in zip(
            current_versions, assert_versions
        ):
            next_version = get_next_dev_version(current_version)

            self.assertEqual(assert_version, next_version)

    def test_get_next_patch_version(self):
        terminal = MagicMock()
        today = datetime.datetime.today()

        filenames = ["pyproject.toml", "CMakeLists.txt"]
        mocks = [
            "pontos.release.helper.PythonVersionCommand",
            "pontos.release.helper.CMakeVersionCommand",
        ]
        current_versions = [
            "20.4.1.dev3",
            f"{str(today.year % 100)}.4.1.dev3",
            f"19.{str(today.month)}.1.dev3",
            f"{str(today.year % 100)}.{str(today.month)}.1",
            "20.6.1",
        ]
        assert_versions = [
            "20.4.1",
            f"{str(today.year % 100)}.4.1",
            f"19.{str(today.month)}.1",
            f"{str(today.year % 100)}.{str(today.month)}.2",
            "20.6.2",
        ]

        with temp_directory(change_into=True) as tmp_path:
            for filename, mock in zip(filenames, mocks):
                for current_version, assert_version in zip(
                    current_versions, assert_versions
                ):
                    proj_file = tmp_path / filename
                    proj_file.touch()
                    with patch(mock) as cmd_mock:
                        # pylint: disable=line-too-long
                        cmd_mock.return_value.get_current_version.return_value = (
                            current_version
                        )

                        release_version = get_next_patch_version(terminal)

                        self.assertEqual(release_version, assert_version)

                    proj_file.unlink()



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=no-member, disallowed-name, line-too-long


import unittest

from pontos.github.api.teams import TeamPrivacy
from pontos.github.models.base import (
    App,
    GitHubModel,
    GitHubModelAttribute,
    Permission,
    Team,
    User,
    dotted_attributes,
)


class DottedAttributesTestCase(unittest.TestCase):
    def test_with_new_class(self):
        class Foo:
            pass

        foo = Foo()
        attrs = {"bar": 123, "hello": "World", "baz": [1, 2, 3]}

        foo = dotted_attributes(foo, attrs)

        self.assertEqual(foo.bar, 123)
        self.assertEqual(foo.baz, [1, 2, 3])
        self.assertEqual(foo.hello, "World")

    def test_with_github_model_attribute(self):
        foo = GitHubModelAttribute()
        attrs = {"bar": 123, "hello": "World", "baz": [1, 2, 3]}

        foo = dotted_attributes(foo, attrs)

        self.assertEqual(foo.bar, 123)
        self.assertEqual(foo.baz, [1, 2, 3])
        self.assertEqual(foo.hello, "World")


class GitHubModelTestCase(unittest.TestCase):
    def test_from_dict(self):
        model = GitHubModel.from_dict(
            {
                "x": 1,
                "y": 2,
                "hello": "World",
                "baz": [1, 2, 3],
                "bar": {"a": "b"},
            }
        )

        self.assertEqual(model.x, 1)
        self.assertEqual(model.y, 2)
        self.assertEqual(model.hello, "World")
        self.assertEqual(model.baz, [1, 2, 3])
        self.assertEqual(model.bar.a, "b")


class UserTestCase(unittest.TestCase):
    def test_from_dict(self):
        data = {
            "login": "greenbone",
            "id": 31986857,
            "node_id": "MDEyOk9yZ2FuaXphdGlvbjMxOTg2ODU3",
            "avatar_url": "https://avatars.githubusercontent.com/u/31986857?v=4",
            "gravatar_id": "",
            "url": "https://api.github.com/users/greenbone",
            "html_url": "https://github.com/greenbone",
            "followers_url": "https://api.github.com/users/greenbone/followers",
            "following_url": "https://api.github.com/users/greenbone/following{/other_user}",
            "gists_url": "https://api.github.com/users/greenbone/gists{/gist_id}",
            "starred_url": "https://api.github.com/users/greenbone/starred{/owner}{/repo}",
            "subscriptions_url": "https://api.github.com/users/greenbone/subscriptions",
            "organizations_url": "https://api.github.com/users/greenbone/orgs",
            "repos_url": "https://api.github.com/users/greenbone/repos",
            "events_url": "https://api.github.com/users/greenbone/events{/privacy}",
            "received_events_url": "https://api.github.com/users/greenbone/received_events",
            "type": "Organization",
            "site_admin": False,
        }

        user = User.from_dict(data)

        self.assertEqual(user.login, "greenbone")
        self.assertEqual(user.id, 31986857)
        self.assertEqual(user.node_id, "MDEyOk9yZ2FuaXphdGlvbjMxOTg2ODU3")
        self.assertEqual(user.gravatar_id, "")
        self.assertEqual(user.url, "https://api.github.com/users/greenbone")
        self.assertEqual(user.html_url, "https://github.com/greenbone")
        self.assertEqual(
            user.followers_url,
            "https://api.github.com/users/greenbone/followers",
        )
        self.assertEqual(
            user.following_url,
            "https://api.github.com/users/greenbone/following{/other_user}",
        )
        self.assertEqual(
            user.gists_url,
            "https://api.github.com/users/greenbone/gists{/gist_id}",
        )
        self.assertEqual(
            user.starred_url,
            "https://api.github.com/users/greenbone/starred{/owner}{/repo}",
        )
        self.assertEqual(
            user.subscriptions_url,
            "https://api.github.com/users/greenbone/subscriptions",
        )
        self.assertEqual(
            user.organizations_url,
            "https://api.github.com/users/greenbone/orgs",
        )
        self.assertEqual(
            user.repos_url, "https://api.github.com/users/greenbone/repos"
        )
        self.assertEqual(
            user.events_url,
            "https://api.github.com/users/greenbone/events{/privacy}",
        )
        self.assertEqual(
            user.received_events_url,
            "https://api.github.com/users/greenbone/received_events",
        )
        self.assertEqual(user.type, "Organization")
        self.assertFalse(user.site_admin)


class TeamTestCase(unittest.TestCase):
    def test_from_dict(self):
        data = {
            "name": "python-gvm-maintainers",
            "id": 3764115,
            "node_id": "MDQ6VGVhbTM3NjQxMTU=",
            "slug": "python-gvm-maintainers",
            "description": "Maintainers of python code at GVM",
            "privacy": "closed",
            "url": "https://api.github.com/organizations/31986857/team/3764115",
            "html_url": "https://github.com/orgs/greenbone/teams/python-gvm-maintainers",
            "members_url": "https://api.github.com/organizations/31986857/team/3764115/members{/member}",
            "repositories_url": "https://api.github.com/organizations/31986857/team/3764115/repos",
            "permission": "pull",
            "parent": None,
        }

        team = Team.from_dict(data)

        self.assertEqual(team.name, "python-gvm-maintainers")
        self.assertEqual(team.id, 3764115)
        self.assertEqual(team.node_id, "MDQ6VGVhbTM3NjQxMTU=")
        self.assertEqual(team.slug, "python-gvm-maintainers")
        self.assertEqual(team.description, "Maintainers of python code at GVM")
        self.assertEqual(team.privacy, TeamPrivacy.CLOSED)
        self.assertEqual(
            team.url,
            "https://api.github.com/organizations/31986857/team/3764115",
        )
        self.assertEqual(
            team.html_url,
            "https://github.com/orgs/greenbone/teams/python-gvm-maintainers",
        )
        self.assertEqual(
            team.members_url,
            "https://api.github.com/organizations/31986857/team/3764115/members{/member}",
        )
        self.assertEqual(
            team.repositories_url,
            "https://api.github.com/organizations/31986857/team/3764115/repos",
        )
        self.assertEqual(team.permission, Permission.PULL)
        self.assertIsNone(team.parent)


class AppTestCase(unittest.TestCase):
    def test_from_dict(self):
        data = {
            "id": 1,
            "slug": "octoapp",
            "node_id": "MDExOkludGVncmF0aW9uMQ==",
            "owner": {
                "login": "github",
                "id": 1,
                "node_id": "MDEyOk9yZ2FuaXphdGlvbjE=",
                "avatar_url": "https://github.com/images/error/octocat_happy.gif",
                "gravatar_id": "",
                "url": "https://api.github.com/orgs/github",
                "html_url": "https://github.com/github",
                "followers_url": "https://api.github.com/users/github/followers",
                "following_url": "https://api.github.com/users/github/following{/other_user}",
                "gists_url": "https://api.github.com/users/github/gists{/gist_id}",
                "starred_url": "https://api.github.com/users/github/starred{/owner}{/repo}",
                "subscriptions_url": "https://api.github.com/users/github/subscriptions",
                "organizations_url": "https://api.github.com/users/github/orgs",
                "repos_url": "https://api.github.com/orgs/github/repos",
                "events_url": "https://api.github.com/orgs/github/events",
                "received_events_url": "https://api.github.com/users/github/received_events",
                "type": "Organization",
                "site_admin": False,
            },
            "name": "Octocat App",
            "description": "",
            "external_url": "https://example.com",
            "html_url": "https://github.com/apps/octoapp",
            "created_at": "2017-07-08T16:18:44-04:00",
            "updated_at": "2017-07-08T16:18:44-04:00",
            "events": ["push", "pull_request"],
        }

        app = App.from_dict(data)

        self.assertEqual(app.id, 1)
        self.assertEqual(app.slug, "octoapp")
        self.assertEqual(app.node_id, "MDExOkludGVncmF0aW9uMQ==")
        self.assertEqual(app.owner.login, "github")
        self.assertEqual(app.owner.id, 1)
        self.assertEqual(app.owner.node_id, "MDEyOk9yZ2FuaXphdGlvbjE=")
        self.assertEqual(
            app.owner.avatar_url,
            "https://github.com/images/error/octocat_happy.gif",
        )
        self.assertEqual(app.owner.gravatar_id, "")
        self.assertEqual(app.owner.url, "https://api.github.com/orgs/github")
        self.assertEqual(app.owner.html_url, "https://github.com/github")
        self.assertEqual(
            app.owner.followers_url,
            "https://api.github.com/users/github/followers",
        )
        self.assertEqual(
            app.owner.following_url,
            "https://api.github.com/users/github/following{/other_user}",
        )
        self.assertEqual(
            app.owner.gists_url,
            "https://api.github.com/users/github/gists{/gist_id}",
        )
        self.assertEqual(
            app.owner.starred_url,
            "https://api.github.com/users/github/starred{/owner}{/repo}",
        )
        self.assertEqual(
            app.owner.subscriptions_url,
            "https://api.github.com/users/github/subscriptions",
        )
        self.assertEqual(
            app.owner.organizations_url,
            "https://api.github.com/users/github/orgs",
        )
        self.assertEqual(
            app.owner.repos_url, "https://api.github.com/orgs/github/repos"
        )
        self.assertEqual(
            app.owner.events_url, "https://api.github.com/orgs/github/events"
        )
        self.assertEqual(app.owner.type, "Organization")
        self.assertFalse(app.owner.site_admin)
        self.assertEqual(app.name, "Octocat App")
        self.assertEqual(app.description, "")
        self.assertEqual(app.external_url, "https://example.com")
        self.assertEqual(app.html_url, "https://github.com/apps/octoapp")
        self.assertEqual(app.created_at, "2017-07-08T16:18:44-04:00")
        self.assertEqual(app.updated_at, "2017-07-08T16:18:44-04:00")
        self.assertEqual(app.events, ["push", "pull_request"])



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from argparse import ArgumentParser, Namespace

from rich.console import Console
from rich.table import Table

from pontos.github.api import GitHubAsyncRESTApi


def add_script_arguments(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--actor", help="Only return workflow runs of this user ID."
    )
    parser.add_argument(
        "--branch", help="Only return workflow runs for a specific branch."
    )
    parser.add_argument(
        "--event",
        help="Only return workflow runs triggered by the event specified. "
        "For example, `push`, `pull_request` or `issue`.",
    )
    parser.add_argument(
        "--status",
        help="Only return workflow runs with the status or conclusion "
        "specified.",
    )
    parser.add_argument(
        "--created",
        help="Only returns workflow runs created within the given date-time "
        "range.",
    )
    parser.add_argument("repository")
    parser.add_argument(
        "workflow",
        help="Workflow ID or workflow file name. For example `main.yml`.",
    )


async def github_script(api: GitHubAsyncRESTApi, args: Namespace) -> int:
    table = Table()
    table.add_column("Name")
    table.add_column("ID")
    table.add_column("URL")
    table.add_column("Branch")
    table.add_column("Event")
    table.add_column("Status")
    table.add_column("Updated")
    table.add_column("Actor")

    count = 0
    async for run in api.workflows.get_workflow_runs(
        args.repository,
        args.workflow,
        actor=args.actor,
        branch=args.branch,
        event=args.event,
        status=args.status,
        created=args.created,
    ):
        table.add_row(
            run.name,
            str(run.id),
            run.html_url,
            run.head_branch,
            run.event,
            run.conclusion,
            run.updated_at,
            run.actor.login,
        )
        count += 1

    console = Console()
    console.print(table)

    print(f"{count} workflow runs.")
    return 0



# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from unittest.mock import patch

from pontos.github.actions.core import ActionIO, Console
from pontos.github.actions.errors import GitHubActionsError
from pontos.testing import temp_directory


@patch("builtins.print")
class ConsoleTestCase(unittest.TestCase):
    def test_start_group(self, print_mock):
        Console.start_group("Foo")
        print_mock.assert_called_once_with("::group::Foo")

    def test_end_group(self, print_mock):
        Console.end_group()
        print_mock.assert_called_once_with("::endgroup::")

    def test_group(self, print_mock):
        with Console.group("Foo"):
            print_mock.assert_called_once_with("::group::Foo")

        print_mock.assert_called_with("::endgroup::")

    def test_warning(self, print_mock):
        Console.warning(
            "foo",
            name="bar",
            line="123",
            end_line="234",
            column="1",
            end_column="2",
            title="Foo Bar",
        )
        print_mock.assert_called_once_with(
            "::warning file=bar,line=123,endLine=234,col=1,endColumn=2,title=Foo Bar::foo"  # pylint: disable=line-too-long
        )

    def test_error(self, print_mock):
        Console.error(
            "foo",
            name="bar",
            line="123",
            end_line="234",
            column="1",
            end_column="2",
            title="Foo Bar",
        )
        print_mock.assert_called_once_with(
            "::error file=bar,line=123,endLine=234,col=1,endColumn=2,title=Foo Bar::foo"  # pylint: disable=line-too-long
        )

    def test_notice(self, print_mock):
        Console.notice(
            "foo",
            name="bar",
            line="123",
            end_line="234",
            column="1",
            end_column="2",
            title="Foo Bar",
        )
        print_mock.assert_called_once_with(
            "::notice file=bar,line=123,endLine=234,col=1,endColumn=2,title=Foo Bar::foo"  # pylint: disable=line-too-long
        )

    def test_log(self, print_mock):
        Console.log("foo")

        print_mock.assert_called_once_with("foo")

    def test_debug(self, print_mock):
        Console.debug("foo")

        print_mock.assert_called_once_with("::debug::foo")


class ActionIOTestCase(unittest.TestCase):
    @patch.dict(
        "os.environ", {"INPUT_FOO": "1234", "INPUT_FOO_BAR": "2345"}, clear=True
    )
    def test_input(self):
        self.assertEqual(ActionIO.input("foo"), "1234")
        self.assertEqual(ActionIO.input("FOO"), "1234")
        self.assertEqual(ActionIO.input("FoO"), "1234")

        self.assertEqual(ActionIO.input("foo bar"), "2345")
        self.assertEqual(ActionIO.input("FOO_BAR"), "2345")
        self.assertEqual(ActionIO.input("FoO BaR"), "2345")

    def test_output(self):
        with temp_directory() as temp_dir:
            file_path = temp_dir / "github.output"

            with patch.dict(
                "os.environ", {"GITHUB_OUTPUT": str(file_path)}, clear=True
            ):
                ActionIO.output("foo", "bar")
                ActionIO.output("lorem", "ipsum")

                output = file_path.read_text(encoding="utf8")

                self.assertEqual(output, "foo=bar\nlorem=ipsum\n")

    def test_output_no_env(self):
        with patch.dict("os.environ", {}, clear=True), self.assertRaises(
            GitHubActionsError
        ):
            ActionIO.output("foo", "bar")

    def test_output_empty_env(self):
        with patch.dict(
            "os.environ", {"GITHUB_OUTPUT": ""}, clear=True
        ), self.assertRaises(GitHubActionsError):
            ActionIO.output("foo", "bar")



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import List, Optional

from pontos.github.models.base import GitHubModel, User

__all__ = (
    "License",
    "Repository",
)


@dataclass
class License(GitHubModel):
    key: str
    name: str
    url: str
    spdx_id: str
    node_id: str
    html_url: Optional[str] = None


@dataclass
class RepositoryPermissions(GitHubModel):
    admin: bool
    push: bool
    pull: bool


@dataclass
class Repository(GitHubModel):
    id: int
    node_id: str
    name: str
    full_name: str
    private: bool
    owner: User
    html_url: str
    description: str
    fork: bool
    url: str
    forks_url: str
    keys_url: str
    collaborators_url: str
    teams_url: str
    hooks_url: str
    issue_events_url: str
    events_url: str
    assignees_url: str
    branches_url: str
    tags_url: str
    blobs_url: str
    git_tags_url: str
    git_refs_url: str
    trees_url: str
    statuses_url: str
    languages_url: str
    stargazers_url: str
    contributors_url: str
    subscribers_url: str
    subscription_url: str
    commits_url: str
    git_commits_url: str
    comments_url: str
    issue_comment_url: str
    contents_url: str
    compare_url: str
    merges_url: str
    archive_url: str
    downloads_url: str
    issues_url: str
    pulls_url: str
    milestones_url: str
    notifications_url: str
    labels_url: str
    releases_url: str
    deployments_url: str
    created_at: Optional[str] = None
    updated_at: Optional[str] = None
    pushed_at: Optional[str] = None
    git_url: Optional[str] = None
    ssh_url: Optional[str] = None
    clone_url: Optional[str] = None
    svn_url: Optional[str] = None
    homepage: Optional[str] = None
    stargazers_count: Optional[int] = None
    watchers_count: Optional[int] = None
    language: Optional[str] = None
    has_issues: Optional[bool] = None
    has_projects: Optional[bool] = None
    has_downloads: Optional[bool] = None
    has_wiki: Optional[bool] = None
    has_pages: Optional[bool] = None
    has_discussions: Optional[bool] = None
    forks_count: Optional[int] = None
    mirror_url: Optional[str] = None
    archived: Optional[bool] = None
    disabled: Optional[bool] = None
    open_issues_count: Optional[int] = None
    is_template: Optional[bool] = None
    topics: Optional[List[str]] = None
    visibility: Optional[str] = None
    forks: Optional[int] = None
    open_issues: Optional[int] = None
    watchers: Optional[int] = None
    default_branch: Optional[str] = None
    allow_forking: Optional[bool] = None
    allow_rebase_merge: Optional[bool] = None
    allow_squash_merge: Optional[bool] = None
    allow_auto_merge: Optional[bool] = None
    delete_branch_on_merge: Optional[bool] = None
    allow_merge_commit: Optional[bool] = None
    web_commit_signoff_required: Optional[bool] = None
    subscribers_count: Optional[int] = None
    network_count: Optional[int] = None
    size: Optional[int] = None
    license: Optional[License] = None
    permissions: Optional[RepositoryPermissions] = None



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# pylint: disable=C0413,W0108

import os
import unittest
from contextlib import redirect_stdout
from io import StringIO
from pathlib import Path
from unittest.mock import MagicMock, call, patch

import httpx

from pontos import changelog, release
from pontos.git import Git
from pontos.release.helper import version
from pontos.testing import temp_git_repository


class ReleaseTestCase(unittest.TestCase):
    def setUp(self) -> None:
        os.environ["GITHUB_TOKEN"] = "foo"
        os.environ["GITHUB_USER"] = "bar"
        self.valid_gh_release_response = (
            '{"zipball_url": "zip", "tarball_url":'
            ' "tar", "upload_url":"upload"}'
        )

    @patch("pontos.release.release.Git", spec=Git)
    @patch("pontos.release.release.Path", spec=Path)
    @patch("pontos.github.api.api.httpx", spec=httpx)
    @patch("pontos.github.api.release.httpx", spec=httpx)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.release.changelog", spec=changelog)
    def test_release_successfully(
        self,
        changelog_mock,
        version_mock,
        requests_mock,
        requests2_mock,
        _path_mock,
        _git_mock,
    ):
        version_mock.main.return_value = (True, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")

        fake_post = MagicMock(spec=httpx.Response).return_value
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        requests_mock.post.return_value = fake_post
        requests2_mock.post.return_value = fake_post

        args = [
            "release",
            "--project",
            "foo",
            "--git-signing-key",
            "123",
            "--release-version",
            "0.0.1",
            "--next-version",
            "0.0.2dev",
        ]

        with temp_git_repository(), redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )
        self.assertTrue(released)

    @patch("pontos.release.release.Git", spec=Git)
    @patch("pontos.release.release.Path", spec=Path)
    @patch("pontos.github.api.api.httpx", spec=httpx)
    @patch("pontos.github.api.release.httpx", spec=httpx)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.release.changelog", spec=changelog)
    def test_release_conventional_commits_successfully(
        self,
        changelog_mock,
        version_mock,
        requests_mock,
        requests2_mock,
        _path_mock,
        _git_mock,
    ):
        version_mock.main.return_value = (True, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")

        fake_post = MagicMock(spec=httpx.Response).return_value
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        requests_mock.post.return_value = fake_post
        requests2_mock.post.return_value = fake_post

        args = [
            "release",
            "--project",
            "foo",
            "--git-signing-key",
            "123",
            "--release-version",
            "1.2.3",
            "-CC",
        ]

        with temp_git_repository(), redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )

        version_mock.main.assert_called_with(
            leave=False, args=["--quiet", "update", "1.2.4", "--develop"]
        )
        self.assertTrue(released)

    @patch("pontos.release.release.Git", spec=Git)
    @patch("pontos.release.release.Path", spec=Path)
    @patch("pontos.github.api.api.httpx", spec=httpx)
    @patch("pontos.github.api.release.httpx", spec=httpx)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.release.changelog", spec=changelog)
    def test_not_release_successfully_when_github_create_release_fails(
        self,
        changelog_mock,
        version_mock,
        requests_mock,
        requests2_mock,
        _path_mock,
        _git_mock,
    ):

        version_mock.main.return_value = (True, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")

        fake_post = MagicMock(spec=httpx.Response).return_value
        fake_post.status_code = 401
        fake_post.text = self.valid_gh_release_response
        fake_post.raise_for_status.side_effect = httpx.HTTPStatusError(
            "Authorization required",
            response=fake_post,
            request=requests_mock.post,
        )
        requests_mock.post.return_value = fake_post
        requests2_mock.post.return_value = fake_post

        args = [
            "release",
            "--release-version",
            "0.0.1",
        ]

        with temp_git_repository(), redirect_stdout(StringIO()):
            git = Git()
            git.add_remote("origin", "https://foo.com/bar.git")

            released = release.main(
                leave=False,
                args=args,
            )
        self.assertFalse(released)

    @patch("pontos.release.release.Git", spec=Git)
    @patch("pontos.release.release.Path", spec=Path)
    @patch("pontos.github.api.api.httpx", spec=httpx)
    @patch("pontos.github.api.release.httpx", spec=httpx)
    @patch("pontos.release.helper.version", spec=version)
    @patch("pontos.release.release.changelog", spec=changelog)
    def test_release_to_specific_git_remote(
        self,
        changelog_mock,
        version_mock,
        requests_mock,
        requests2_mock,
        _path_mock,
        git_mock,
    ):
        version_mock.main.return_value = (True, "MyProject.conf")
        changelog_mock.update.return_value = ("updated", "changelog")

        fake_post = MagicMock(spec=httpx.Response).return_value
        fake_post.status_code = 201
        fake_post.text = self.valid_gh_release_response
        requests_mock.post.return_value = fake_post
        requests2_mock.post.return_value = fake_post

        args = [
            "release",
            "--project",
            "foo",
            "--release-version",
            "0.0.1",
            "--next-version",
            "0.0.2.dev1",
            "--git-remote-name",
            "upstream",
            "--git-signing-key",
            "1234",
        ]

        with temp_git_repository(), redirect_stdout(StringIO()):
            released = release.main(
                leave=False,
                args=args,
            )
        self.assertTrue(released)

        git_mock.return_value.push.assert_called_with(
            follow_tags=True, remote="upstream"
        )
        # git_mock.return_value.add.assert_called_with()
        git_mock.return_value.add.assert_has_calls(
            [
                call("MyProject.conf"),
                call("*__version__.py"),
                call("CHANGELOG.md"),
            ]
        )
        git_mock.return_value.commit.assert_called_with(
            "Automatic adjustments after release\n\n"
            "* Update to version 0.0.2.dev1\n"
            "* Add empty changelog after 0.0.1",
            verify=False,
            gpg_signing_key="1234",
        )



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This script creates a new repository with default settings
"""

import shutil
from argparse import ArgumentParser, Namespace
from typing import Union

from pontos.git import Git, MergeStrategy
from pontos.github.api import GitHubAsyncRESTApi
from pontos.github.api.repositories import GitIgnoreTemplate, LicenseType
from pontos.github.models.base import Permission
from pontos.testing import temp_directory

TEMPLATES = {
    "python": "https://github.com/greenbone/python-project-template.git",
    "go": "https://github.com/greenbone/go-project-template.git",
}

GITIGNORE = {"python": GitIgnoreTemplate.PYTHON, "go": GitIgnoreTemplate.GO}


def license_type(value: Union[str, LicenseType]) -> LicenseType:
    if isinstance(value, LicenseType):
        return value

    return LicenseType(value.lower())


def possible_license_types() -> str:
    return ", ".join(
        [
            LicenseType.GNU_GENERAL_PUBLIC_LICENSE_2_0.value,
            LicenseType.GNU_GENERAL_PUBLIC_LICENSE_3_0.value,
            LicenseType.GNU_AFFERO_GENERAL_PUBLIC_LICENSE_3_0.value,
        ]
    )


def add_script_arguments(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--template",
        choices=("python", "go"),
        help="Use template repo as base for the new repository.",
    )
    parser.add_argument(
        "--team",
        help="Team that should have admin access to the repository.",
    )
    parser.add_argument(
        "--license",
        help=f"License to choose for the repo: {possible_license_types()}. "
        "Default: %(default)s.",
        type=license_type,
        default=LicenseType.GNU_AFFERO_GENERAL_PUBLIC_LICENSE_3_0.value,
    )
    parser.add_argument(
        "--visibility",
        choices=("public", "private"),
        default="private",
        help="Visibility of the repository. Default: %(default)s.",
    )
    parser.add_argument("--description", help="Description of the repository.")
    parser.add_argument("name", help="Repository to create.")
    parser.add_argument(
        "organization",
        nargs="?",
        default="greenbone",
        help="Organization to create the repo in. Default: %(default)s.",
    )


async def github_script(api: GitHubAsyncRESTApi, args: Namespace) -> int:
    organization = args.organization
    repository = args.name

    team_id = None
    if args.team:
        team = await api.teams.get(organization, args.team)
        team_id = team.id

    with temp_directory() as temp_dir:
        git = Git()

        if args.template:
            git.clone(TEMPLATES[args.template], temp_dir, remote="template")
            dot_git_dir = temp_dir / ".git"
            shutil.rmtree(dot_git_dir)

            git.cwd = temp_dir
            git.init()
            git.add(".")
            git.commit(f"Starting commit from {args.template} template")
        else:
            git.cwd = temp_dir
            git.init()

        private = True if args.visibility == "private" else False

        repo = await api.repositories.create(
            organization,
            repository,
            private=private,
            has_projects=False,
            has_wiki=False,
            allow_merge_commit=True,
            allow_auto_merge=True,
            allow_rebase_merge=True,
            allow_squash_merge=True,
            allow_update_branch=True,
            delete_branch_on_merge=True,
            is_template=False,
            license_template=args.license,
            description=args.description,
            auto_init=True,
            gitignore_template=GITIGNORE.get(args.template),
            team_id=team_id,
        )

        repo_url = repo.ssh_url

        git.add_remote("upstream", repo_url)
        git.fetch("upstream")

        if args.template:
            git.rebase("upstream/main", strategy=MergeStrategy.ORT_OURS)
        else:
            git.checkout("main", start_point="upstream/main")

        if args.team:
            await api.teams.update_permission(
                organization, team.slug, repository, Permission.ADMIN
            )

            code_owners_file = temp_dir / ".github" / "CODEOWNERS"
            code_owners_file.write_text(
                f"# default reviewers\n*\t@{organization}/{team.slug}\n"
            )
            git.add(code_owners_file)
            git.commit("Adjust CODEOWNERS file")

        git.push(remote="upstream", force=True)

        await api.branches.update_protection_rules(
            f"{organization}/{repository}",
            "main",
            require_branches_to_be_up_to_date=True,
            require_code_owner_reviews=True,
            required_approving_review_count=1,
            required_conversation_resolution=True,
            dismiss_stale_reviews=True,
            allow_force_pushes=False,
            allow_deletions=False,
            restrictions_users=[],
        )



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from enum import Enum
from inspect import isclass
from typing import Any, Dict, List, Optional, Type, Union, get_type_hints

try:
    from typing import get_args, get_origin
except ImportError:
    from typing_extensions import get_args, get_origin

__all__ = (
    "App",
    "GitHubModel",
    "User",
    "Team",
)


def dotted_attributes(obj: Any, data: Dict[str, Any]) -> Any:
    """
    Set dotted attributes on an object

    Example:
        .. code-block:: python

            class Foo:
                '''Some class'''

            foo = Foo()
            attrs = {"bar": 123, "baz": 456}

            foo = dotted_attributes(foo, attrs)
            print(foo.bar, foo.baz)
    """
    for key, value in data.items():
        if isinstance(value, dict):
            default = None if hasattr(obj, key) else GitHubModelAttribute()
            prop = getattr(obj, key, default)
            value = dotted_attributes(prop, value)

        setattr(obj, key, value)

    return obj


class GitHubModelAttribute:
    """
    A utility class to allow setting attributes
    """


@dataclass(init=False)
class GitHubModel:
    """
    Base class for all GitHub models
    """

    @staticmethod
    def _get_value_from_model_field_cls(
        model_field_cls: Type[Any], value: Any
    ) -> Any:
        if isclass(model_field_cls) and issubclass(
            model_field_cls, GitHubModel
        ):
            value = model_field_cls.from_dict(value)
        elif (
            get_origin(model_field_cls) == Union
            or get_origin(model_field_cls) == list
        ):
            # it should be an Optional field
            model_field_cls = get_args(model_field_cls)[0]
            value = GitHubModel._get_value_from_model_field_cls(
                model_field_cls, value
            )
        else:
            if isinstance(value, dict):
                value = model_field_cls(**value)
            else:
                value = model_field_cls(value)
        return value

    @staticmethod
    def _get_value(model_field_cls: Type[Any], value: Any) -> Any:
        if model_field_cls:
            value = GitHubModel._get_value_from_model_field_cls(
                model_field_cls, value
            )
        return value

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """
        Create a model from a dict

        Example:
            .. code-block:: python

            model = GitHubModel.from_dict({
                "id": 123,
                "node_id": "abcde",
                "created_at": "2017-07-08T16:18:44-04:00",
                "updated_at": "2017-07-08T16:18:44-04:00",
            })
        """
        kwargs = {}
        additional_attrs = {}
        type_hints = get_type_hints(cls)
        for name, value in data.items():
            if isinstance(value, list):
                model_field_cls = type_hints.get(name)
                value = [cls._get_value(model_field_cls, v) for v in value]
            elif value is not None:
                model_field_cls = type_hints.get(name)
                value = cls._get_value(model_field_cls, value)

            if name in type_hints:
                kwargs[name] = value
            else:
                additional_attrs[name] = value

        instance = cls(**kwargs)
        dotted_attributes(instance, additional_attrs)
        return instance


@dataclass
class User(GitHubModel):
    login: str
    id: int
    node_id: str
    avatar_url: str
    gravatar_id: str
    url: str
    html_url: str
    followers_url: str
    following_url: str
    gists_url: str
    starred_url: str
    subscriptions_url: str
    organizations_url: str
    repos_url: str
    events_url: str
    received_events_url: str
    type: str
    site_admin: bool


class TeamPrivacy(Enum):
    SECRET = "secret"
    CLOSED = "closed"


class TeamRole(Enum):
    MEMBER = "member"
    MAINTAINER = "maintainer"


class Permission(Enum):
    PULL = "pull"
    PUSH = "push"
    TRIAGE = "triage"
    MAINTAIN = "maintain"
    ADMIN = "admin"


@dataclass
class Team(GitHubModel):
    id: int
    node_id: str
    url: str
    html_url: str
    name: str
    slug: str
    description: str
    privacy: TeamPrivacy
    permission: Permission
    members_url: str
    repositories_url: str
    parent: Optional["Team"] = None


@dataclass
class App(GitHubModel):
    id: int
    slug: str
    node_id: str
    owner: User
    name: str
    description: str
    external_url: str
    html_url: str
    created_at: str
    updated_at: str
    events: List[str]



# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import sys
from pathlib import Path
from typing import Optional, Tuple

from packaging.version import InvalidVersion, Version

from pontos import version
from pontos.git import Git, GitError
from pontos.terminal import Terminal
from pontos.version import CMakeVersionCommand, PythonVersionCommand
from pontos.version.helper import VersionError

DEFAULT_TIMEOUT = 1000
DEFAULT_CHUNK_SIZE = 4096


def commit_files(
    git: Git,
    filename: str,
    commit_msg: str,
    *,
    git_signing_key: str = "",
    changelog: bool = False,
):
    """Add files to staged and commit staged files.

    Args:
        git: A Git instance
        filename: The filename of file to add and commit
        commit_msg: The commit message for the commit
        git_signing_key: The signing key to sign this commit
        changelog: Add CHANGELOG.md file

    """
    git.add(filename)

    try:
        git.add("*__version__.py")
    except GitError:
        pass

    if changelog:
        git.add("CHANGELOG.md")
    if git_signing_key:
        git.commit(commit_msg, verify=False, gpg_signing_key=git_signing_key)
    else:
        git.commit(commit_msg, verify=False)


def calculate_calendar_version(terminal: Terminal) -> str:
    """find the correct next calendar version by checking latest version and
    the today's date"""

    current_version_str: str = get_current_version(terminal)
    current_version = Version(current_version_str)

    today = datetime.date.today()

    if (
        current_version.major < today.year % 100
        or current_version.minor < today.month
    ):
        release_version = Version(
            f"{str(today.year  % 100)}.{str(today.month)}.0"
        )
        return str(release_version)
    elif (
        current_version.major == today.year % 100
        and current_version.minor == today.month
    ):
        if current_version.dev is None:
            release_version = Version(
                f"{str(today.year  % 100)}.{str(today.month)}."
                f"{str(current_version.micro + 1)}"
            )
        else:
            release_version = Version(
                f"{str(today.year  % 100)}.{str(today.month)}."
                f"{str(current_version.micro)}"
            )
        return str(release_version)
    else:
        terminal.error(
            f"'{str(current_version)}' is higher than "
            f"'{str(today.year  % 100)}.{str(today.month)}'."
        )
        sys.exit(1)


def get_current_version(terminal: Terminal) -> Optional[str]:
    """Get the current Version from a pyproject.toml or
    a CMakeLists.txt file"""

    available_cmds = [
        ("CMakeLists.txt", CMakeVersionCommand),
        ("pyproject.toml", PythonVersionCommand),
    ]
    for file_name, cmd in available_cmds:
        project_definition_path = Path.cwd() / file_name
        if project_definition_path.exists():
            terminal.ok(f"Found {file_name} project definition file.")
            current_version: str = cmd().get_current_version()
            return current_version

    terminal.error("No project settings file found")
    return None


def get_last_release_version() -> Optional[str]:
    """Get the last released Version from git.

    Returns:
        Last released git-tag if tags were found
        or None
    """

    git_interface = Git()
    tag_list = git_interface.list_tags()
    return tag_list[-1] if tag_list else None


def get_next_patch_version(terminal: Terminal) -> str:
    """find the correct next patch version by checking latest version"""

    current_version_str: str = get_current_version(terminal)
    current_version = Version(current_version_str)

    if current_version.dev is not None:
        release_version = Version(
            f"{str(current_version.major)}."
            f"{str(current_version.minor)}."
            f"{str(current_version.micro)}"
        )
    else:
        release_version = Version(
            f"{str(current_version.major)}."
            f"{str(current_version.minor)}."
            f"{str(current_version.micro + 1)}"
        )

    return str(release_version)


def get_next_dev_version(release_version: str) -> str:
    """Get the next dev Version from a valid version"""
    # will be a dev1 version
    try:
        release_version_obj = Version(release_version)
        next_version_obj = Version(
            f"{str(release_version_obj.major)}."
            f"{str(release_version_obj.minor)}."
            f"{str(release_version_obj.micro + 1)}"
        )
        return str(next_version_obj)
    except InvalidVersion as e:
        raise (VersionError(e)) from None


def get_git_repository_name(
    remote: str = "origin",
) -> str:
    """Get the git repository name

    Arguments:
        remote: the remote to look up the name (str) default: origin

    Returns:
        project name
    """

    ret = Git().remote_url(remote)
    return ret.rsplit("/", maxsplit=1)[-1].replace(".git", "").strip()


def find_signing_key(terminal: Terminal) -> str:
    """Find the signing key in the config

    Arguments:
        terminal: The terminal for console output

    Returns:
        git signing key or empty string
    """

    try:
        return Git().config("user.signingkey").strip()
    except GitError as e:
        # The command `git config user.signingkey` returns
        # return code 1 if no key is set.
        # So we will return empty string ...
        if e.returncode == 1:
            terminal.warning("No signing key found.")
        return ""


def update_version(
    terminal: Terminal, to: str, *, develop: bool = False
) -> Tuple[bool, str]:
    """Use pontos-version to update the version.

    Arguments:
        to: The version (str) that will be set
        develop: Wether to set version to develop or not (bool)

    Returns:
       executed: True if successfully executed, False else
       filename: The filename of the project definition
    """
    args = ["--quiet"]
    args.append("update")
    args.append(to)
    if develop:
        args.append("--develop")
    executed, filename = version.main(leave=False, args=args)

    if not executed:
        if filename == "":
            terminal.error("No project definition found.")
        else:
            terminal.error(f"Unable to update version {to} in {filename}")

    return executed, filename



# -*- coding: utf-8 -*-
# pontos/release/release.py
# Copyright (C) 2020-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import subprocess
import sys
from argparse import ArgumentParser, FileType, Namespace
from pathlib import Path
from typing import Tuple

from pontos.terminal.null import NullTerminal
from pontos.terminal.rich import RichTerminal

from .prepare import prepare
from .release import release
from .sign import sign


def parse_args(args) -> Tuple[str, str, Namespace]:
    parser = ArgumentParser(
        description="Release handling utility.",
        prog="pontos-release",
    )

    parser.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help="Don't print messages to the terminal",
    )

    parser.add_argument(
        "--log-file",
        dest="log_file",
        type=str,
        help="Activate logging using the given file path",
    )

    subparsers = parser.add_subparsers(
        title="subcommands",
        description="valid subcommands",
        help="additional help",
        dest="command",
        required=True,
    )

    prepare_parser = subparsers.add_parser("prepare")
    prepare_parser.set_defaults(func=prepare)
    version_group = prepare_parser.add_mutually_exclusive_group(required=True)
    version_group.add_argument(
        "--release-version",
        help="Will release changelog as version. Must be PEP 440 compliant",
    )
    version_group.add_argument(
        "--calendar",
        help=(
            "Automatically calculate calendar release version, from current"
            " version and date."
        ),
        action="store_true",
    )
    version_group.add_argument(
        "--patch",
        help=("Release next patch version: " "e.g. x.x.3 -> x.x.4"),
        action="store_true",
    )

    prepare_parser.add_argument(
        "--git-signing-key",
        help="The key to sign the commits and tag for a release",
    )
    prepare_parser.add_argument(
        "--git-tag-prefix",
        default="v",
        help="Prefix for git tag versions. Default: %(default)s",
    )
    prepare_parser.add_argument(
        "--changelog",
        help=(
            "The CHANGELOG file path, defaults "
            "to CHANGELOG.md in the repository root directory"
        ),
    )
    prepare_parser.add_argument(
        "--space",
        default="greenbone",
        help="User/Team name in github",
    )
    prepare_parser.add_argument(
        "--project",
        help="The github project",
    )
    prepare_parser.add_argument(
        "--conventional-commits",
        "-CC",
        help=(
            "Wether to use conventional commits and create "
            "the changelog directly from the git log"
        ),
        action="store_true",
    )
    prepare_parser.add_argument(
        "--conventional-commits-config",
        dest="cc_config",
        default=Path("changelog.toml"),
        type=FileType("r"),
        help="Conventional commits config file (toml), including conventions.",
    )

    release_parser = subparsers.add_parser("release")
    release_parser.set_defaults(func=release)
    release_parser.add_argument(
        "--release-version",
        help=(
            "Will release changelog as version. Must be PEP 440 compliant. "
            "default: lookup version in project definition."
        ),
    )

    release_parser.add_argument(
        "--next-version",
        help=(
            "Sets the next PEP 440 compliant version in project definition "
            "after the release. default: set to next dev version"
        ),
    )

    release_parser.add_argument(
        "--git-remote-name",
        help="The git remote name to push the commits and tag to",
    )
    release_parser.add_argument(
        "--git-tag-prefix",
        default="v",
        help="Prefix for git tag versions. Default: %(default)s",
    )
    release_parser.add_argument(
        "--git-signing-key",
        help="The key to sign the commits and tag for a release",
    )
    release_parser.add_argument(
        "--project",
        help="The github project",
    )
    release_parser.add_argument(
        "--space",
        default="greenbone",
        help="User/Team name in github",
    )
    release_parser.add_argument(
        "--changelog",
        help=(
            "The CHANGELOG file path, defaults "
            "to CHANGELOG.md in the repository root directory"
        ),
    )
    release_parser.add_argument(
        "--conventional-commits",
        "-CC",
        help=(
            "Wether to use conventional commits and create "
            "the changelog directly from the git log"
        ),
        action="store_true",
    )

    sign_parser = subparsers.add_parser("sign")
    sign_parser.set_defaults(func=sign)
    sign_parser.add_argument(
        "--signing-key",
        default="0ED1E580",
        help="The key to sign zip, tarballs of a release. Default %(default)s.",
    )
    sign_parser.add_argument(
        "--release-version",
        help="Will release changelog as version. Must be PEP 440 compliant.",
    )
    sign_parser.add_argument(
        "--git-tag-prefix",
        default="v",
        help="Prefix for git tag versions. Default: %(default)s",
    )
    sign_parser.add_argument(
        "--project",
        help="The github project",
    )
    sign_parser.add_argument(
        "--space",
        default="greenbone",
        help="user/team name in github",
    )

    sign_parser.add_argument(
        "--passphrase",
        help=(
            "Use gpg in a headless mode e.g. for "
            "the CI and use this passphrase for signing."
        ),
    )
    sign_parser.add_argument(
        "--dry-run", action="store_true", help="Do not upload signed files."
    )
    parsed_args = parser.parse_args(args)
    token = os.getenv("GITHUB_TOKEN") if not args else "TOKEN"
    user = os.getenv("GITHUB_USER") if not args else "USER"
    return user, token, parsed_args


def main(
    leave: bool = True,
    args=None,
):
    username, token, parsed_args = parse_args(args)
    if parsed_args.quiet:
        term = NullTerminal()
    else:
        term = RichTerminal()

    term.bold_info(f"pontos-release => {parsed_args.func.__name__}")

    with term.indent():
        try:
            if not parsed_args.func(
                term,
                parsed_args,
                username=username,
                token=token,
            ):
                return sys.exit(1) if leave else False
        except subprocess.CalledProcessError as e:
            if not "--passphrase" in e.cmd:
                term.error(f'Could not run command "{e.cmd}".')
            else:
                term.error("Headless signing failed.")

            term.print(f"Error was: {e.stderr}")
            sys.exit(1)

    return sys.exit(0) if leave else True


if __name__ == "__main__":
    main()



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from contextlib import AbstractAsyncContextManager
from types import TracebackType
from typing import Any, AsyncContextManager, AsyncIterator, Dict, Optional, Type

import httpx

from pontos.github.api.helper import (
    DEFAULT_GITHUB_API_URL,
    DEFAULT_TIMEOUT_CONFIG,
    JSON,
    JSON_OBJECT,
    _get_next_url,
)
from pontos.github.models.base import GitHubModel

Headers = Dict[str, str]
Params = Dict[str, str]


class GitHubAsyncRESTClient(AbstractAsyncContextManager):
    """
    A client for calling the GitHub REST API asynchronously

    Should be used as an async context manager

    Example:
        .. code-block:: python

        org = "foo"
        async with GitHubAsyncRESTClient(token) as client:
            response = await client.get(f"/orgs/{org}/repos")
    """

    def __init__(
        self,
        token: Optional[str] = None,
        url: Optional[str] = DEFAULT_GITHUB_API_URL,
        *,
        timeout: Optional[httpx.Timeout] = DEFAULT_TIMEOUT_CONFIG,
    ) -> None:
        self.token = token
        self.url = url
        self._client = httpx.AsyncClient(timeout=timeout, http2=True)

    def _request_headers(
        self, *, content_type: Optional[str] = None
    ) -> Headers:
        """
        Get the default request headers
        """
        headers = {
            "Accept": "application/vnd.github.v3+json",
        }
        if self.token:
            headers["Authorization"] = f"token {self.token}"
        if content_type:
            headers["Content-Type"] = content_type

        return headers

    def _request_kwargs(
        self, *, json: Optional[JSON] = None, content: Optional[Any] = None
    ) -> Dict[str, str]:
        """
        Get the default request arguments
        """
        kwargs = {}
        if json is not None:
            kwargs["json"] = json
        if content is not None:
            kwargs["content"] = content
        return kwargs

    def _request_api_url(self, api: str) -> str:
        return f"{self.url}{api}"

    def _request_url(self, api_or_url: str) -> str:
        return (
            api_or_url
            if api_or_url.startswith("https")
            else self._request_api_url(api_or_url)
        )

    async def get(
        self,
        api: str,
        *,
        params: Optional[Params] = None,
    ) -> httpx.Response:
        """
        Get request to a GitHub API

        Args:
            api: API path to use for the get request
            params: Optional params to use for the get request
        """
        url = self._request_url(api)
        headers = self._request_headers()
        kwargs = self._request_kwargs()
        return await self._client.get(
            url,
            headers=headers,
            params=params,
            follow_redirects=True,
            **kwargs,
        )

    async def get_all(
        self,
        api: str,
        *,
        params: Optional[Params] = None,
    ) -> AsyncIterator[httpx.Response]:
        """
        Get paginated content of a get GitHub API request

        Args:
            api: API path to use for the get request
            params: Optional params to use for the get request
        """
        response = await self.get(api, params=params)

        yield response

        next_url = _get_next_url(response)

        while next_url:
            response = await self.get(next_url, params=params)

            yield response

            next_url = _get_next_url(response)

    async def delete(
        self, api: str, *, params: Optional[Params] = None
    ) -> httpx.Response:
        """
        Delete request to a GitHub API

        Args:
            api: API path to use for the delete request
            params: Optional params to use for the delete request
        """
        headers = self._request_headers()
        url = self._request_url(api)
        return await self._client.delete(url, params=params, headers=headers)

    async def post(
        self,
        api: str,
        *,
        data: Optional[JSON] = None,
        params: Optional[Params] = None,
        content: Optional[str] = None,
        content_type: Optional[str] = None,
    ) -> httpx.Response:
        """
        Post request to a GitHub API

        Args:
            api: API path to use for the post request
            params: Optional params to use for the post request
            data: Optional data to include in the post request
        """
        headers = self._request_headers(content_type=content_type)
        url = self._request_url(api)
        return await self._client.post(
            url, params=params, headers=headers, json=data, content=content
        )

    async def put(
        self,
        api: str,
        *,
        data: Optional[JSON] = None,
        params: Optional[Params] = None,
        content: Optional[str] = None,
        content_type: Optional[str] = None,
    ) -> httpx.Response:
        """
        Put request to a GitHub API

        Args:
            api: API path to use for the put request
            params: Optional params to use for the put request
            data: Optional data to include in the put request
        """
        headers = self._request_headers(content_type=content_type)
        url = self._request_url(api)
        return await self._client.put(
            url, params=params, headers=headers, json=data, content=content
        )

    async def patch(
        self,
        api: str,
        *,
        data: Optional[JSON] = None,
        params: Optional[Params] = None,
        content: Optional[str] = None,
        content_type: Optional[str] = None,
    ) -> httpx.Response:
        """
        Patch request to a GitHub API

        Args:
            api: API path to use for the patch request
            params: Optional params to use for the patch request
            data: Optional data to include in the patch request
        """
        headers = self._request_headers(content_type=content_type)
        url = self._request_url(api)
        return await self._client.patch(
            url, params=params, headers=headers, json=data, content=content
        )

    def stream(self, api: str) -> AsyncContextManager[httpx.Response]:
        """
        Stream data from a GitHub API

        Args:
            api: API path to use for the post request
        """
        headers = self._request_headers()
        url = self._request_url(api)
        return self._client.stream(
            "GET", url, headers=headers, follow_redirects=True
        )

    async def __aenter__(self) -> "GitHubAsyncRESTClient":
        await self._client.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Optional[bool]:
        return await self._client.__aexit__(exc_type, exc_value, traceback)


class GitHubAsyncREST:
    """
    Base class for GitHub asynchronous REST classes
    """

    def __init__(self, client: GitHubAsyncRESTClient):
        self._client = client

    async def _get_paged_items(
        self,
        api: str,
        name: str,
        model_cls: Type[GitHubModel],
        *,
        params: Optional[Params] = None,
    ) -> AsyncIterator[GitHubModel]:
        """
        Internal method to get the paged items information from different REST
        URLs.
        """
        if not params:
            params = {}

        params["per_page"] = "100"  # max number

        async for response in self._client.get_all(api, params=params):
            response.raise_for_status()
            data: JSON_OBJECT = response.json()
            for item in data.get(name, []):
                yield model_cls.from_dict(item)



# -*- coding: utf-8 -*-
# Copyright (C) 2021-2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from unittest.mock import MagicMock, patch

from pontos import changelog
from pontos.changelog.conventional_commits import (
    ChangelogBuilderError,
    parse_args,
)
from pontos.git import Git
from pontos.git.git import TagSort
from pontos.testing import temp_directory


@dataclass
class StdOutput:
    stdout: bytes


class ConventionalCommitsTestCase(unittest.TestCase):
    @patch("pontos.changelog.conventional_commits.Git", spec=Git)
    def test_changelog_builder(self, git_mock):
        terminal = MagicMock()

        today = datetime.today().strftime("%Y-%m-%d")

        own_path = Path(__file__).absolute().parent
        config_toml = own_path / "changelog.toml"
        release_version = "0.0.2"
        output = f"v{release_version}.md"

        git_mock.return_value.list_tags.return_value = ["v0.0.1"]
        git_mock.return_value.log.return_value = [
            "1234567 Add: foo bar",
            "8abcdef Add: bar baz",
            "8abcd3f Add bar baz",
            "8abcd3d Adding bar baz",
            "1337abc Change: bar to baz",
            "42a42a4 Remove: foo bar again",
            "fedcba8 Test: bar baz testing",
            "dead901 Refactor: bar baz ref",
            "fedcba8 Fix: bar baz fixing",
            "d0c4d0c Doc: bar baz documenting",
        ]

        changelog_builder = changelog.ChangelogBuilder(
            terminal=terminal,
            current_version="0.0.1",
            next_version=release_version,
            space="foo",
            project="bar",
            config=config_toml,
        )
        expected_output = f"""# Changelog

All notable changes to this project will be documented in this file.

## [0.0.2] - {today}

## Added
* foo bar [1234567](https://github.com/foo/bar/commit/1234567)
* bar baz [8abcdef](https://github.com/foo/bar/commit/8abcdef)

## Removed
* foo bar again [42a42a4](https://github.com/foo/bar/commit/42a42a4)

## Changed
* bar to baz [1337abc](https://github.com/foo/bar/commit/1337abc)

## Bug Fixes
* bar baz fixing [fedcba8](https://github.com/foo/bar/commit/fedcba8)

## Documentation
* bar baz documenting [d0c4d0c](https://github.com/foo/bar/commit/d0c4d0c)

## Refactor
* bar baz ref [dead901](https://github.com/foo/bar/commit/dead901)

## Testing
* bar baz testing [fedcba8](https://github.com/foo/bar/commit/fedcba8)

[0.0.2]: https://github.com/foo/bar/compare/0.0.1...0.0.2"""

        with temp_directory(change_into=True):
            output_file = changelog_builder.create_changelog_file(output)
            self.assertEqual(
                output_file.read_text(encoding="utf-8"), expected_output
            )

        git_mock.return_value.list_tags.assert_called_with(sort=TagSort.VERSION)
        git_mock.return_value.log.assert_called_with(
            "v0.0.1..HEAD", oneline=True
        )

    @patch("pontos.changelog.conventional_commits.Git")
    def test_changelog_builder_no_commits(self, git_mock):
        terminal = MagicMock()

        own_path = Path(__file__).absolute().parent
        config_toml = own_path / "changelog.toml"
        release_version = "0.0.2"
        output = f"v{release_version}.md"

        git_mock.return_value.list_tags.return_value = ["v0.0.1"]
        git_mock.return_value.log.return_value = []

        changelog_builder = changelog.ChangelogBuilder(
            terminal=terminal,
            current_version="0.0.1",
            next_version=release_version,
            space="foo",
            project="bar",
            config=config_toml,
        )

        with self.assertRaises(ChangelogBuilderError):
            output_file = changelog_builder.create_changelog_file(output)

            self.assertIsNone(output_file)

        git_mock.return_value.log.assert_called_with(
            "v0.0.1..HEAD", oneline=True
        )

    @patch("pontos.changelog.conventional_commits.Git")
    def test_changelog_builder_no_tag(self, git_mock):
        terminal = MagicMock()

        today = datetime.today().strftime("%Y-%m-%d")

        own_path = Path(__file__).absolute().parent
        config_toml = own_path / "changelog.toml"
        release_version = "0.0.2"
        output = f"v{release_version}.md"
        git_mock.return_value.list_tags.return_value = []
        git_mock.return_value.log.return_value = [
            "1234567 Add: foo bar",
            "8abcdef Add: bar baz",
            "8abcd3f Add bar baz",
            "8abcd3d Adding bar baz",
            "1337abc Change: bar to baz",
            "42a42a4 Remove: foo bar again",
            "fedcba8 Test: bar baz testing",
            "dead901 Refactor: bar baz ref",
            "fedcba8 Fix: bar baz fixing",
            "d0c4d0c Doc: bar baz documenting",
        ]

        changelog_builder = changelog.ChangelogBuilder(
            terminal=terminal,
            current_version="0.0.1",
            next_version=release_version,
            project="bar",
            space="foo",
            config=config_toml,
        )
        expected_output = f"""# Changelog

All notable changes to this project will be documented in this file.

## [0.0.2] - {today}

## Added
* foo bar [1234567](https://github.com/foo/bar/commit/1234567)
* bar baz [8abcdef](https://github.com/foo/bar/commit/8abcdef)

## Removed
* foo bar again [42a42a4](https://github.com/foo/bar/commit/42a42a4)

## Changed
* bar to baz [1337abc](https://github.com/foo/bar/commit/1337abc)

## Bug Fixes
* bar baz fixing [fedcba8](https://github.com/foo/bar/commit/fedcba8)

## Documentation
* bar baz documenting [d0c4d0c](https://github.com/foo/bar/commit/d0c4d0c)

## Refactor
* bar baz ref [dead901](https://github.com/foo/bar/commit/dead901)

## Testing
* bar baz testing [fedcba8](https://github.com/foo/bar/commit/fedcba8)

[0.0.2]: https://github.com/foo/bar/compare/0.0.1...0.0.2"""
        with temp_directory(change_into=True):
            output_file = changelog_builder.create_changelog_file(output)
            self.assertEqual(
                output_file.read_text(encoding="utf-8"), expected_output
            )

        git_mock.return_value.list_tags.assert_called_with(sort=TagSort.VERSION)
        git_mock.return_value.log.assert_called_with(oneline=True)

    @patch("pontos.changelog.conventional_commits.Git")
    def test_changelog_builder_no_conventional_commits(self, git_mock):
        terminal = MagicMock()

        own_path = Path(__file__).absolute().parent
        config_toml = own_path / "changelog.toml"
        release_version = "0.0.2"
        output = f"v{release_version}.md"
        git_mock.return_value.list_tags.return_value = ["v0.0.1"]
        git_mock.return_value.log.return_value = [
            "1234567 foo bar",
            "8abcdef bar baz",
        ]
        changelog_builder = changelog.ChangelogBuilder(
            terminal=terminal,
            current_version="0.0.1",
            next_version=release_version,
            space="foo",
            project="bar",
            config=config_toml,
        )

        with self.assertRaises(ChangelogBuilderError):
            output_file = changelog_builder.create_changelog_file(output)

            self.assertIsNone(output_file)


class ParseArgsTestCase(unittest.TestCase):
    def test_parse_args(self):
        parsed_args = parse_args(
            ["-q", "--log-file", "blah", "--project", "urghs"]
        )

        self.assertTrue(parsed_args.quiet)
        self.assertEqual(parsed_args.log_file, "blah")
        self.assertEqual(parsed_args.project, "urghs")



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Load and run Pontos GitHub Scripts

A Pontos GitHub Script is a Python module that has a github_script and
optionally a add_script_arguments functions. These functions should have the
following signatures:

.. code-block:: python

    async def github_script(api: GitHubAsyncRESTApi, args: Namespace) -> int:

    def add_script_arguments(parser: ArgumentParser) -> None:

Instead of an async coroutine a sync function is also possible:

.. code-block:: python

    def github_script(api: GitHubRESTApi, args: Namespace) -> int:

Example:

.. code-block:: python
    :caption: Example Python module containing a Pontos GitHub Script

    def add_script_arguments(parser: ArgumentParser) -> None:
        parser.add_argument("repository")

    async def github_script(api: GitHubAsyncRESTApi, args: Namespace) -> int:
        repo_data = api.repository.get(args.repository)
        print(repo_data["description"])
        return 0
"""

import json
import sys
from argparse import ArgumentParser

import httpx

from pontos.errors import PontosError
from pontos.github.script.load import (
    load_script,
    run_add_arguments_function,
    run_github_script_function,
)
from pontos.github.script.parser import create_parser


def main():
    """
    CLI function to run a Pontos GitHub Script
    """
    parser = create_parser()
    known_args, _ = parser.parse_known_args()
    try:
        with load_script(known_args.script) as module:
            # using a child parser allows for adding --help including the script
            # arguments
            child_parser = ArgumentParser(parents=[parser])
            run_add_arguments_function(module, child_parser)
            args = child_parser.parse_args()
            token = args.token
            timeout = args.timeout

            retval = run_github_script_function(module, token, timeout, args)

        sys.exit(retval)
    except KeyboardInterrupt:
        sys.exit(1)
    except httpx.HTTPStatusError as e:
        try:
            error = e.response.json()
            message = error.get("message")
            print("Error:", message, file=sys.stderr)
        except json.JSONDecodeError:
            # not a json response
            print(e, file=sys.stderr)

        sys.exit(1)
    except PontosError as e:
        print(e, file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright (C) 2022 Greenbone Networks GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import List, Optional

from pontos.github.models.base import App, GitHubModel, Team, User

__all__ = (
    "BranchProtection",
    "BranchProtectionFeature",
    "BypassPullRequestAllowances",
    "DismissalRestrictions",
    "RequiredPullRequestReviews",
    "RequiredStatusChecks",
    "Restrictions",
    "StatusCheck",
)


@dataclass
class DismissalRestrictions(GitHubModel):
    url: str
    users_url: str
    teams_url: str
    users: List[User]
    teams: List[Team]
    apps: List[App]


@dataclass
class BypassPullRequestAllowances(GitHubModel):
    users: List[User]
    teams: List[Team]
    apps: List[App]


@dataclass
class RequiredPullRequestReviews(GitHubModel):
    url: str
    dismiss_stale_reviews: bool
    require_code_owner_reviews: bool
    required_approving_review_count: int
    require_last_push_approval: bool
    dismissal_restrictions: Optional[DismissalRestrictions] = None
    bypass_pull_request_allowances: Optional[BypassPullRequestAllowances] = None


@dataclass
class StatusCheck(GitHubModel):
    context: str
    app_id: Optional[int] = None


@dataclass
class RequiredStatusChecks(GitHubModel):
    url: str
    strict: bool
    checks: List[StatusCheck]
    enforcement_level: Optional[str] = None


@dataclass
class Restrictions(GitHubModel):
    url: str
    users_url: str
    teams_url: str
    apps_url: str
    users: List[User]
    teams: List[Team]
    apps: List[App]


@dataclass
class BranchProtectionFeature(GitHubModel):
    enabled: bool
    url: Optional[str] = None


@dataclass
class BranchProtection(GitHubModel):
    url: str
    required_status_checks: Optional[RequiredStatusChecks] = None
    required_pull_request_reviews: Optional[RequiredPullRequestReviews] = None
    restrictions: Optional[Restrictions] = None
    enforce_admins: Optional[BranchProtectionFeature] = None
    required_linear_history: Optional[BranchProtectionFeature] = None
    allow_force_pushes: Optional[BranchProtectionFeature] = None
    allow_deletions: Optional[BranchProtectionFeature] = None
    block_creations: Optional[BranchProtectionFeature] = None
    required_conversation_resolution: Optional[BranchProtectionFeature] = None
    lock_branch: Optional[BranchProtectionFeature] = None
    allow_fork_syncing: Optional[BranchProtectionFeature] = None
    required_signatures: Optional[BranchProtectionFeature] = None
